import pytest

from .base import (parse_res, V01, app, app_notest,
                   get_features_cached)
from freediscovery.exceptions import WrongParameter

# ============================================================================#
#
#                     LSI
#
# ============================================================================#

def test_api_lsi(app):
    dsid, pars, _ = get_features_cached(app)
    method = V01 + "/feature-extraction/{}".format(dsid)
    res = app.get(method)
    assert res.status_code == 200
    data = parse_res(res)
    method = V01 + "/lsi/"
    res = app.get_check(method, data=dict(parent_id=dsid))

    lsi_pars = dict(n_components=101, parent_id=dsid)
    method = V01 + "/lsi/"
    data = app.post_check(method, json=lsi_pars)
    assert sorted(data.keys()) == ['explained_variance', 'id']
    lid = data['id']

    # checking again that we can load all the lsi models
    method = V01 + "/lsi/"
    data = app.get(method, data=dict(parent_id=dsid,))

    method = V01 + "/lsi/{}".format(lid)
    data = app.get_check(method)
    for key, vals in lsi_pars.items():
        assert vals == data[key]

    assert sorted(data.keys()) == sorted(["n_components", "parent_id"])

    for key in data.keys():
        assert data[key] == lsi_pars[key]

def test_api_lsi_custom_names(app):
    dsid, pars, _ = get_features_cached(app)
    mid_orig = 'tyds'
    lsi_pars = dict(n_components=2, parent_id=dsid, id=mid_orig)
    data = app.post_check(V01 + "/lsi/", json=lsi_pars)
    assert data['id'] == mid_orig
    method = V01 + "/lsi/{}".format(mid_orig)
    data = app.get_check(method)

    with pytest.raises(WrongParameter):
        data = app.post_check(V01 + "/lsi/", json=lsi_pars)

    with pytest.raises(WrongParameter):
        lsi_pars['id'] = '?+d'
        data = app.post_check(V01 + "/lsi/", json=lsi_pars)



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import warnings

import numpy as np
from sklearn.preprocessing import normalize
from sklearn.decomposition import TruncatedSVD
from sklearn.utils import check_array
from sklearn.utils.extmath import safe_sparse_dot


def _compute_lsi_dimensionality(n_components, n_samples, n_features,
                                alpha=0.33):
    """ Reduce the number of LSI components for small datasets """
    n_components_samples = min(n_components, alpha*n_samples)
    msg = []
    if n_components_samples < n_components:
        msg.append(('The ingested dataset has only {} documents while {} LSI '
                    'components were requested; '
                    'decreasing the number of LSI components: ')
                   .format(n_samples, n_components))

    n_components_feature = int(min(n_components, alpha*n_features))
    if n_components_feature < n_components:
        msg.append(('The vocabulary in the ingested dataset has '
                    'only {} words (or n-grams) while {} LSI '
                    'components were requested; '
                    'decreasing the number of LSI components: ')
                   .format(n_features, n_components))
    n_components_opt = int(min(n_components_samples, n_components_feature))
    n_components_opt = max(5, n_components_opt)
    if n_components_opt < n_components:
        msg.append('Decreasing n_components from {} to {}'
                   .format(n_components, n_components_opt))
    if msg:
        msg = '\n'.join(msg)
        warnings.warn(msg)
    return n_components_opt


class _TruncatedSVD_LSI(TruncatedSVD):
    """
    A patch of `sklearn.decomposition.TruncatedSVD` to include whitening
    (`scikit-learn/scikit-learn#7832)`
    """

    def transform_lsi(self, X):
        """ LSI transform, normalized by the inverse of the eigen values"""
        X = check_array(X, accept_sparse='csr')
        return safe_sparse_dot(X, self.components_.T).dot(
                    np.diag(1./self.singular_values_[:self.n_components]))

    def transform_lsi_norm(self, X):
        Y = self.transform_lsi(X)
        normalize(Y, copy=False)
        return Y



# -*- coding: utf-8 -*-

import os.path
from unittest import SkipTest
import numpy as np
import pytest

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.tests.run_suite import check_cache


def fd_setup(**fe_options):
    basename = os.path.dirname(__file__)
    cache_dir = check_cache()
    data_dir = os.path.join(basename, "..", "..", "data", "ds_001", "raw")
    n_features = 110000
    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(n_features=n_features, use_hashing=True,
                    stop_words='english',
                    **fe_options)
    fe.ingest(data_dir, file_pattern='.*\d.txt')
    return cache_dir, uuid, fe.filenames_, fe


@pytest.mark.parametrize('method, options, fe_options',
                         [['simhash', {'distance': 3}, {}],
                          ['simhash', {'distance': 10}, {}],
                          ['i-match', {}, {}]])
def test_dup_detection(method, options, fe_options):
    if method == 'simhash':
        try:
            import simhash
        except ImportError:
            raise SkipTest
    from freediscovery.engine.near_duplicates import _DuplicateDetectionWrapper
    cache_dir, uuid, filenames, fe = fd_setup(**fe_options)

    dd = _DuplicateDetectionWrapper(cache_dir=cache_dir, parent_id=uuid)
    dd.fit(method=method)
    cluster_id = dd.query(**options)
    # cannot have more cluster_id than elements in the dataset
    assert len(np.unique(cluster_id)) <= len(np.unique(filenames))



# Authors: Roman Yurchak
#
# License: BSD 3 clause

from __future__ import division

import numpy as np
from freediscovery.cluster.utils import _dbscan_unique2noisy

from sklearn.exceptions import UndefinedMetricWarning

# Information retrieval specific metrics
# See scikit learn metrics for more general ones


# Categorization Metrics


def recall_at_k_score(y_true, y_pred, k):
    """
    Recall after retrieving k documents from the collections

    Parameters
    ----------
    y_true : ndarray [n_samples]
      array of integer classes
    y_pred : ndarray [n_samples]
      array of float predicted scores
    k : {int, float}
      the threashold either float in [0., 1.] or int in [0, n_samples]

    Returns
    -------
    score : float
      the recall at k score
    """
    N_docs = len(y_true)
    if isinstance(k, float):
        k = int(N_docs*k)
    elif isinstance(k, int):
        pass
    else:
        raise TypeError('Provided k with type {} must be int or float'
                        .format(type(k)))

    if len(y_true) != len(y_pred):
        raise ValueError('len(y_true)={} != len(y_pred)={}'
                         .format(len(y_true) != len(y_pred)))

    index_sorted = np.argsort(y_pred)[::-1]
    recall_curve = y_true[index_sorted].cumsum()/y_true.sum()
    # make sure the element 0 is always 0
    recall_curve = np.hstack((np.zeros(1), recall_curve))
    return recall_curve[k]


# Clustering Metrics
def ratio_duplicates_score(x, y):
    """
    Given cluster labels x and y, compute the relative error
    between the number of duplicates in x vs the one in y.
    """
    if x.shape != y.shape:
        raise ValueError
    N = len(x)
    N_dups_x = N - len(np.unique(x))
    N_dups_y = N - len(np.unique(y))

    return 1 - abs(N_dups_x - N_dups_y)/max(N_dups_x, N_dups_y)


def f1_same_duplicates_score(x, y):
    """
    Given cluster labels x and y, compute the f1 score
    that the same elements are marked as duplicates
    """
    import warnings
    from sklearn.metrics import f1_score

    if x.shape != y.shape:
        raise ValueError
    x_dup = _dbscan_unique2noisy(x)
    x_dup[x_dup > -1] = 1  # duplicates
    x_dup[x_dup == -1] = 0  # not duplicates
    y_dup = _dbscan_unique2noisy(y)
    y_dup[y_dup > -1] = 1  # duplicates
    y_dup[y_dup == -1] = 0  # not duplicates

    x_dup = np.abs(x_dup)
    y_dup = np.abs(y_dup)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UndefinedMetricWarning)
        score = f1_score(x_dup, y_dup)

    return score


def mean_duplicates_count_score(x, y):
    """ Score based on the number of duplicates for sample k,
    averaged over samples."""
    from freediscovery.utils import _count_duplicates
    x_count = _count_duplicates(x)
    y_count = _count_duplicates(y)
    mask = (x_count > 1) | (y_count > 1)  # select only duplicates
    x_count = x_count[mask]
    y_count = y_count[mask]

    score = 1 - np.abs(x_count - y_count)/np.maximum(x_count, y_count)
    return np.mean(score)


def seuclidean_dist2cosine_sim(x):
    """Given an squared euclidean distance on L2 normalized data,
    convert it to the cosine similarity

    Warning: this function would give completely wrong results if,
      * the euclidean distance is not squared
      * the data is initially not L2 normalized
    """
    return 1 - x/2.


def cosine2jaccard_similarity(s_cos):
    """ Given a cosine similarity on L2 normalized data,
    compute the jaccard similarity

    Parameters
    ----------
    s_cos : {float, ndarray}
      the cosine similarity

    Returns
    -------
    s_jac : {float, ndarray}
      the Jaccard similarity
    """
    return s_cos / (2 - s_cos)


def jaccard2cosine_similarity(s_jac):
    """ Given a cosine similarity on L2 normalized data,
    compute the jaccard similarity

    Parameters
    ----------
    s_jac : {float, ndarray}
      the cosine similarity

    Returns
    -------
    s_cos : {float, ndarray}
      the Jaccard similarity
    """
    return 2*s_jac / (1 + s_jac)


def _normalize_similarity(x, metric='cosine', inverse=False):
    """Given a similarity score, normalize it to the
    [0, 1] range

    Parameters
    ----------
    x : {float, ndarray}
      the similarity score
    metric : str
      the metric used (one of 'cosine', 'jaccard')
    inverse : bool
      perform the inverse de-normalization operation
    """
    if metric == 'cosine':
        # cosine similarity can be in the [-1, 1] range
        if not inverse:
            return (x + 1)/2
        else:
            return 2*x - 1
    elif metric == 'jaccard':
        # jaccard similarity could potenitally be in the [-1/3, 1] range
        # when using the cosine2jaccard_similarity function
        if not inverse:
            return (3*x + 1)/4.
        else:
            return (4*x - 1)/3.
    else:
        raise ValueError


def _scale_cosine_similarity(x, metric='cosine', inverse=False):
    """ Given a cosine similarity on L2 normalized data,
    appriximately convert it to Jaccard similarity, and/or
    normalize it to the [0, 1] interval

    Parameters
    ----------
    x : {float, ndarray}
      the cosine similarity value
    metric : str
      the conversion to apply one of ['cosine', 'jaccard']
    inverse : bool
      perform the inverse de-normalization operation
    """
    valid_metrics = ['cosine', 'jaccard', 'cosine_norm', 'jaccard_norm',
                     'cosine-positive']
    if metric not in valid_metrics:
        raise ValueError('metric {} not supported, must be in {}'
                         .format(metric, valid_metrics))
    if metric == 'cosine':
        return x
    elif metric == 'cosine-positive':
        if isinstance(x, (int, float)):
            return max(x, 0.0)
        else:
            return np.fmax(x, 0.0)

    if metric.startswith('jaccard'):
        if not inverse:
            x = cosine2jaccard_similarity(x)
        else:
            x = jaccard2cosine_similarity(x)

    if metric.endswith('norm'):
        x = _normalize_similarity(x, metric=metric.split('_')[0],
                                  inverse=inverse)

    return x


def categorization_score(idx_ref, Y_ref, idx, Y):
    """ Calculate the efficiency scores """
    # This function should be deprecated
    # An equivalent functionally should be achieved with a
    # more general freediscovery.metrics module
    import warnings
    from sklearn.metrics import precision_score, recall_score, f1_score
    from sklearn.metrics import roc_auc_score, average_precision_score
    threshold = 0.0

    idx = np.asarray(idx, dtype='int')
    idx_ref = np.asarray(idx_ref, dtype='int')
    Y = np.asarray(Y)
    Y_ref = np.asarray(Y_ref)

    idx_out = np.intersect1d(idx_ref, idx)
    if not len(idx_out):
        return {"recall_score": -1, "precision_score": -1, 'f1': -1,
                'auc_roc': -1, 'average_precision': -1}

    # sort values by index
    order_ref = idx_ref.argsort()
    idx_ref = idx_ref[order_ref]
    Y_ref = Y_ref[order_ref]

    order = idx.argsort()
    idx = idx[order]
    Y = Y[order]

    # find indices that are in both the reference and the test dataset
    mask_ref = np.in1d(idx_ref, idx_out)
    mask = np.in1d(idx, idx_out)

    Y_ref = Y_ref[mask_ref]
    Y = Y[mask]
    Y_bin = (Y > threshold)

    n_classes = len(np.unique(Y_ref))

    if n_classes != 2:
        opts = {"average": 'micro'}
    else:
        opts = {}

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UndefinedMetricWarning)

        m_recall_score = recall_score(Y_ref, Y_bin, **opts)
        m_precision_score = precision_score(Y_ref, Y_bin, **opts)
        m_f1_score = f1_score(Y_ref, Y_bin, **opts)
    if n_classes == 2:
        m_roc_auc = roc_auc_score(Y_ref, Y)
        m_average_precision = average_precision_score(Y_ref, Y)
        m_recall_at_20p = recall_at_k_score(Y_ref, Y, k=0.2)
    else:
        # not defined for non binary categorization
        m_roc_auc = np.nan
        m_average_precision = np.nan
        m_recall_at_20p = np.nan

    return {"recall": m_recall_score, "precision": m_precision_score,
            "f1": m_f1_score, 'roc_auc': m_roc_auc,
            'average_precision': m_average_precision,
            'recall_at_20p': m_recall_at_20p}



"""
Data Ingestion
==============

An example illustrating the data ingestion in FreeDiscovery
"""

from __future__ import print_function

import requests
import pandas as pd
import json
import os.path

pd.options.display.float_format = '{:,.3f}'.format
pd.options.display.expand_frame_repr = False

dataset_name = "treclegal09_2k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

###############################################################################
#
#  0. Load the test dataset

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()

# To use a custom dataset, simply specify the following variables
# create a custom dataset definition for ingestion
data_dir = input_ds['metadata']['data_dir']
dataset_definition = [{'document_id': row['document_id'],
                       'file_path': os.path.join(data_dir, row['file_path'])} \
                               for row in input_ds['dataset']]


###############################################################################
#
# 1.a Load dataset and initalize feature extraction
url = BASE_URL + '/feature-extraction'
print(" POST", url)
res = requests.post(url, json={'use_hashing': True}).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))

print("\n1.b Start feature extraction")

url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
res = requests.post(url, json={'dataset_definition': dataset_definition})


###############################################################################
#
# 2 check the parameters of the extracted features
url = BASE_URL + '/feature-extraction/{}'.format(dsid)
print(' GET', url)
res = requests.get(url).json()

print('\n'.join(['     - {}: {}'.format(key, val)
      for key, val in res.items() if "filenames" not in key]))

###############################################################################
#
# 3. Examine the id mapping

method = BASE_URL + "/feature-extraction/{}/id-mapping".format(dsid)
print('\n GET', method)
data = {'data': [{'internal_id': row['internal_id']} for row in input_ds['dataset'][:3]]}
print('   DATA', json.dumps(data))
res = requests.post(method, json=data).json()

print(' Response:')
print('  ', json.dumps(res, indent=4))

###############################################################################
#
# 4 Delete the extracted features

url = BASE_URL + '/feature-extraction/{}'.format(dsid)
print(" DELETE", url)
requests.delete(url)



# -*- coding: utf-8 -*-

import os.path

from flask import Flask, jsonify, json
from flask.testing import FlaskClient
from flask_apispec import FlaskApiSpec

from apispec import APISpec

from .resources import (FeaturesApi, FeaturesApiElement,
                        FeaturesApiElementMappingNested,
                        FeaturesApiAppend, FeaturesApiRemove,
                        ModelsApi, ModelsApiElement, ModelsApiPredict,
                        LsiApi, LsiApiElement,
                        ClusteringApiElement, KmeanClusteringApi,
                        BirchClusteringApi, DBSCANClusteringApi,
                        DupDetectionApi, DupDetectionApiElement,
                        ExampleDatasetApi,
                        MetricsCategorizationApiElement,
                        MetricsClusteringApiElement,
                        MetricsDupDetectionApiElement,
                        EmailThreadingApi, EmailThreadingApiElement,
                        SearchApi,
                        CustomStopWordsApi, CustomStopWordsLoadApi,
                        )
from .info import ServerInfoApi


if os.environ.get('DOCKER', None) is not None:
    from ..externals.flask_compress import Compress

    compress = Compress()
else:
    compress = None

# class CatchAll(Resource):
#
#    def get(self, url):
#        print("accessing", url)
#
#    def post(self, url):
#        print("accessing", url)


class TestClient(FlaskClient):
    # Addresses https://github.com/pallets/flask/issues/2176
    def open(self, *args, **kwargs):
        if 'json' in kwargs:
            kwargs['data'] = json.dumps(kwargs.pop('json'))
            kwargs['content_type'] = 'application/json'
        return super(TestClient, self).open(*args, **kwargs)


def fd_app(cache_dir, config=None):
    """ API app for FreeDiscovery """

    if not os.path.exists(cache_dir):
        raise ValueError('Cache_dir {} does not exist!'.format(cache_dir))
    app = Flask('freediscovery_api')
    app.url_map.strict_slashes = False
    app.config.update(
         {'APISPEC_SPEC': APISpec(title='FreeDiscovery',
                                  version='v0',
                                  plugins=['apispec.ext.marshmallow']),
          'APISPEC_SWAGGER_URL': '/openapi-specs.json',
          'APISPEC_SWAGGER_UI_URL': '/swagger-ui.html'})

    app.test_client_class = TestClient
    if compress is not None:
        compress.init_app(app)

    docs = FlaskApiSpec(app)

    #app.config['DEBUG'] = False

    ServerInfoApi._fd_config = config
    app.add_url_rule('/', view_func=ServerInfoApi.as_view('ServerInfoApi'))

    # Actually setup the Api resource routing here
    for resource_el, url in [
         (ExampleDatasetApi               , "/example-dataset/<name>"),
         (FeaturesApi                     , "/feature-extraction")            ,
         (FeaturesApiElement              , '/feature-extraction/<dsid>')     ,
         (FeaturesApiElementMappingNested , '/feature-extraction/<dsid>/id-mapping'),
         (FeaturesApiAppend               , '/feature-extraction/<dsid>/append') ,
         (FeaturesApiRemove               , '/feature-extraction/<dsid>/delete') ,
         (ModelsApi                       , '/categorization/')               ,
         (ModelsApiElement                , '/categorization/<mid>')          ,
         (ModelsApiPredict                , '/categorization/<mid>/predict')  ,
         (LsiApi                          , '/lsi/')                          ,
         (LsiApiElement                   , '/lsi/<mid>')                     ,
         (KmeanClusteringApi              , '/clustering/k-mean/')            ,
         (BirchClusteringApi              , '/clustering/birch')              ,
         (DBSCANClusteringApi             , '/clustering/dbscan')             ,
         (ClusteringApiElement            , '/clustering/<method>/<mid>')     ,
         (DupDetectionApi                 , '/duplicate-detection/')          ,
         (DupDetectionApiElement          , '/duplicate-detection/<mid>')     ,
         (MetricsCategorizationApiElement , '/metrics/categorization')        ,
         (MetricsClusteringApiElement     , '/metrics/clustering')            ,
         (MetricsDupDetectionApiElement   , '/metrics/duplicate-detection')   ,
         (EmailThreadingApi               , '/email-threading/')              ,
         (EmailThreadingApiElement        , '/email-threading/<mid>')         ,
         (SearchApi                       , '/search/')                       ,
         (CustomStopWordsApi              , '/stop-words/')                   ,
         (CustomStopWordsLoadApi          , '/stop-words/<name>')             ,
         #(CatchAll                       , "/<url>")
                             ]:
        # monkeypatching, not great
        resource_el._cache_dir = cache_dir
        resource_el._random_seed = os.environ.get('FREEDISCOVERY_RANDOM_SEED', 42)
        resource_el.methods = ['GET', 'POST', 'DELETE']
        #api.add_resource(resource_el, '/api/v0' + url, strict_slashes=False)

        ressource_name = resource_el.__name__
        app.add_url_rule('/api/v0' + url,
                         view_func=resource_el.as_view(ressource_name))

        if ressource_name not in ["EmailThreadingApi"]:
            # the above two cases fail due to recursion limit
            docs.register(resource_el, endpoint=ressource_name)

    @app.errorhandler(500)
    def handle_error(error):
        # response = jsonify(error.to_dict())
        response = jsonify({'messages': str(error)})
        response.status_code = 500
        return response

    @app.errorhandler(404)
    def handle_404_error(error):
        response = jsonify({"messages": str(error)})
        response.status_code = 404
        return response

    @app.errorhandler(400)
    def handle_400_error(error):
        response = jsonify({"messages": str(error)})
        response.status_code = 400
        return response

    @app.errorhandler(422)
    def handle_422_error(error):
        # marshmallow error
        response = jsonify(error.data)
        response.status_code = 422
        return response

    return app



"""
Tests for the birch clustering algorithm taken from
  sklearn/cluster/tests/test_birch.py

patched with scikit-learn/scikit-learn#8808

Ensures that the patched version of Birch included in
FreeDiscovery version is still compatible with upstream.
"""

from scipy import sparse
import numpy as np

from sklearn.cluster.tests.common import generate_clustered_data
from freediscovery.cluster.birch import Birch
from sklearn.cluster.hierarchical import AgglomerativeClustering
from sklearn.datasets import make_blobs
from sklearn.linear_model import ElasticNet
from sklearn.metrics import pairwise_distances_argmin, v_measure_score

from sklearn.utils.testing import assert_greater_equal
from sklearn.utils.testing import assert_equal
from sklearn.utils.testing import assert_greater
from sklearn.utils.testing import assert_almost_equal
from sklearn.utils.testing import assert_array_equal
from sklearn.utils.testing import assert_array_almost_equal
from sklearn.utils.testing import assert_raises
from sklearn.utils.testing import assert_warns


def test_n_samples_leaves_roots():
    # Sanity check for the number of samples in leaves and roots
    X, y = make_blobs(n_samples=10)
    brc = Birch()
    brc.fit(X)
    n_samples_root = sum([sc.n_samples_ for sc in brc.root_.subclusters_])
    n_samples_leaves = sum([sc.n_samples_ for leaf in brc._get_leaves()
                            for sc in leaf.subclusters_])
    assert_equal(n_samples_leaves, X.shape[0])
    assert_equal(n_samples_root, X.shape[0])


def test_birch_predict():
    # Test the predict method predicts the nearest centroid.
    rng = np.random.RandomState(0)
    X = generate_clustered_data(n_clusters=3, n_features=3,
                                n_samples_per_cluster=10)

    # n_samples * n_samples_per_cluster
    shuffle_indices = np.arange(30)
    rng.shuffle(shuffle_indices)
    X_shuffle = X[shuffle_indices, :]
    brc = Birch(n_clusters=4, threshold=1.)
    brc.fit(X_shuffle)
    centroids = brc.subcluster_centers_
    assert_array_equal(brc.labels_, brc.predict(X_shuffle))
    nearest_centroid = pairwise_distances_argmin(X_shuffle, centroids)
    assert_almost_equal(v_measure_score(nearest_centroid, brc.labels_), 1.0)


def test_n_clusters():
    # Test that n_clusters param works properly
    X, y = make_blobs(n_samples=100, centers=10)
    brc1 = Birch(n_clusters=10)
    brc1.fit(X)
    assert_greater(len(brc1.subcluster_centers_), 10)
    assert_equal(len(np.unique(brc1.labels_)), 10)

    # Test that n_clusters = Agglomerative Clustering gives
    # the same results.
    gc = AgglomerativeClustering(n_clusters=10)
    brc2 = Birch(n_clusters=gc)
    brc2.fit(X)
    assert_array_equal(brc1.subcluster_labels_, brc2.subcluster_labels_)
    assert_array_equal(brc1.labels_, brc2.labels_)

    # Test that the wrong global clustering step raises an Error.
    clf = ElasticNet()
    brc3 = Birch(n_clusters=clf)
    assert_raises(ValueError, brc3.fit, X)

    # Test that a small number of clusters raises a warning.
    brc4 = Birch(threshold=10000.)
    assert_warns(UserWarning, brc4.fit, X)


def test_sparse_X():
    # Test that sparse and dense data give same results
    X, y = make_blobs(n_samples=100, centers=10)
    brc = Birch(n_clusters=10)
    brc.fit(X)

    csr = sparse.csr_matrix(X)
    brc_sparse = Birch(n_clusters=10)
    brc_sparse.fit(csr)

    assert_array_equal(brc.labels_, brc_sparse.labels_)
    assert_array_almost_equal(brc.subcluster_centers_,
                              brc_sparse.subcluster_centers_)


def check_branching_factor(node, branching_factor):
    subclusters = node.subclusters_
    assert_greater_equal(branching_factor, len(subclusters))
    for cluster in subclusters:
        if cluster.child_:
            check_branching_factor(cluster.child_, branching_factor)


def test_branching_factor():
    # Test that nodes have at max branching_factor number of subclusters
    X, y = make_blobs()
    branching_factor = 9

    # Purposefully set a low threshold to maximize the subclusters.
    brc = Birch(n_clusters=None, branching_factor=branching_factor,
                threshold=0.01)
    brc.fit(X)
    check_branching_factor(brc.root_, branching_factor)
    brc = Birch(n_clusters=3, branching_factor=branching_factor,
                threshold=0.01)
    brc.fit(X)
    check_branching_factor(brc.root_, branching_factor)

    # Raises error when branching_factor is set to one.
    brc = Birch(n_clusters=None, branching_factor=1, threshold=0.01)
    assert_raises(ValueError, brc.fit, X)


def check_threshold(birch_instance, threshold):
    """Use the leaf linked list for traversal"""
    current_leaf = birch_instance.dummy_leaf_.next_leaf_
    while current_leaf:
        subclusters = current_leaf.subclusters_
        for sc in subclusters:
            assert_greater_equal(threshold, sc.radius)
        current_leaf = current_leaf.next_leaf_


def test_threshold():
    # Test that the leaf subclusters have a threshold lesser than radius
    X, y = make_blobs(n_samples=80, centers=4)
    brc = Birch(threshold=0.5, n_clusters=None)
    brc.fit(X)
    check_threshold(brc, 0.5)

    brc = Birch(threshold=5.0, n_clusters=None)
    brc.fit(X)
    check_threshold(brc, 5.)


def test_birch_hierarchy():
    X, y = make_blobs(random_state=40)
    brc = Birch(n_clusters=None, branching_factor=5,
                compute_sample_indices=True)
    brc.fit(X)

    # make sure that leave nodes contain all the samples
    n_leaves = 1
    sample_id = []
    current_leaf = brc.dummy_leaf_.next_leaf_
    while current_leaf:
        subclusters = current_leaf.subclusters_
        for sc in subclusters:
            assert sc.n_samples_ == len(sc.samples_id_)
            sample_id += sc.samples_id_
        current_leaf = current_leaf.next_leaf_
        n_leaves += 1
    assert_array_equal(np.sort(sample_id), np.arange(X.shape[0]))

    # Verify that the resulting hierarchical tree is deeper than 1 level
    # (i.e. subclusters of the root node are nor tree leaves )
    assert len(brc.root_.subclusters_) < n_leaves

    # Make sure that subclusters of the root_ node contain all the samples
    sample_id = []
    for sc in brc.root_.subclusters_:
        sample_id += sc.samples_id_
        assert sc.n_samples_ == len(sc.samples_id_)
    assert_array_equal(np.sort(sample_id), np.arange(X.shape[0]))

    # Pick a sample at random and make sure that reported samples_id_
    # matches with the subcluster the sample is closest to
    document_id = 45
    document_in_subcluster = []
    distance_to_centroid = []
    for sc in brc.root_.subclusters_:
        centroid = X[sc.samples_id_, :].mean(axis=0)
        distance_to_centroid.append(((X[[document_id]] - centroid)**2).sum())
        document_in_subcluster.append(document_id in sc.samples_id_)

    assert np.argmin(distance_to_centroid) == \
        np.nonzero(document_in_subcluster)[0][0]

    # Make sure that we can recompute labels from tree leaves
    labels2 = np.zeros(X.shape[0], dtype=int)
    cluster_id = 0
    for current_leaf in brc._get_leaves():
        subclusters = current_leaf.subclusters_
        for sc in subclusters:
            labels2[list(sc.samples_id_)] = cluster_id
            cluster_id += 1

    assert np.unique(brc.labels_).shape == np.unique(labels2).shape
    # The two methods yield approximately equal labels
    assert v_measure_score(brc.labels_, labels2) > 0.95



# Authors: Roman Yurchak
#
# License: BSD 3 clause

from sklearn.metrics import pairwise_distances


class Search(object):
    """ (Semantic) search in a document collection

    Parameters
    ----------
    vectorizer : {CountVectorizer, HashingVectorizer}
      the (fitted) vectorizer that was used extract tokens from the
      document collection
    tfidf : {TfidfTransformer, SmartTfidfTransfomer}
      the (fitted) IDF transformer used to weight and normalize the
      bag of word/n-gram features
    lsi : TruncatedSVD
      (optional) an LSI model fitted on the vectorised document-term matrix
      If provided this corresponds to a semantic search, default=None
    """
    def __init__(self, vectorizer, tfidf, lsi=None):
        self.vectorizer = vectorizer
        self.tfidf = tfidf
        self.lsi = lsi
        self._fit_X = None

    def fit(self, X):
        """
        Fit using a document term matrix (optionally in the LSI space)

        Parameters
        ----------
        X : ndarray
          the sparse document-terms arrays (if lsi was not used) or
          dense documents / lsi terms array (if lsi was provided)
        """
        self._fit_X = X

    def search(self, text, metric='cosine'):
        """
        Perform the search operation

        Parameters
        ----------
        text : str
          the search query text
        metric : str
          the output metric to use
        """
        from .lsi import _TruncatedSVD_LSI

        if self._fit_X is None:
            raise ValueError('Estomator must be fitted before '
                             'using the search method!')
        q_vect = self.vectorizer.transform([text])
        q_vect = self.tfidf.transform(q_vect)

        if self.lsi is not None:
            # this is a hack need to be rewritten
            if isinstance(self.lsi, _TruncatedSVD_LSI):
                q_lsi = self.lsi.transform_lsi_norm(q_vect)
            else:  # a regular TruncatedSVD object
                q_lsi = self.lsi.transform(q_vect)
            q = q_lsi
        else:
            q = q_vect

        scores = self._compute_score(q, self._fit_X, metric)

        return scores

    def search_id(self, internal_id, metric='cosine'):
        """
        Perform the search operation

        Parameters
        ----------
        internal_id : int
          the internal_id of the document used as a search query
        metric : str
          the output metric to use
        """
        if self._fit_X is None:
            raise ValueError('Estomator must be fitted before '
                             'using the search method!')

        X = self._fit_X

        q = X[[internal_id], :]

        scores = self._compute_score(q, self._fit_X, metric)

        return scores

    @staticmethod
    def _compute_score(q, X, metric):
        """ Internal method to compute the scores """

        from .metrics import _scale_cosine_similarity

        dist = pairwise_distances(q, X, 'cosine')
        dist = dist[0]

        scores = 1 - dist

        scores = _scale_cosine_similarity(scores, metric=metric)

        return scores



# -*- coding: utf-8 -*-

import pytest
import itertools
from numpy.testing import assert_almost_equal

from .. import fd_app

from .base import parse_res, V01, app, app_notest


@pytest.mark.parametrize('metrics',
                         itertools.combinations(['precision', 'recall', 'f1',
                                                 'roc_auc', 'average_precision'], 3))
def test_categorization_metrics(app, metrics):
    metrics = list(metrics)
    url = V01 + '/metrics/categorization'
    labels = ['negative', 'positive']


    y_true = [{'category': labels[key],
               'document_id': idx} for idx, key in\
                      enumerate([0, 0, 0, 1, 1, 0, 1, 0, 1])]
    y_pred = [{'scores': [{'category': labels[key],
                           'score': 1.}],
               'document_id': idx} for idx, key in\
                      enumerate([0, 0, 1, 1, 1, 0, 1, 1, 1])]

    pars = {'y_true': y_true, 'y_pred': y_pred, 'metrics': metrics}
    data = app.post_check(url, json=pars)
    assert sorted(data.keys()) == sorted(metrics)
    for key in metrics:
        assert data[key] > 0.5
        assert data[key] <= 1.0



@pytest.mark.parametrize('metrics',
                         itertools.combinations(['adjusted_rand', 'adjusted_mutual_info', 'v_measure'], 2))
def test_clustering_metrics(app, metrics):
    metrics = list(metrics)
    url = V01 + '/metrics/clustering'
    labels_true = [0, 0, 1, 2]
    labels_pred = [0, 0, 1, 1]

    pars = {'labels_true': labels_true, 'labels_pred': labels_pred, 'metrics': metrics}
    data = app.post_check(url, json=pars)

    assert sorted(data.keys()) == sorted(metrics)
    if 'adjusted_rand' in metrics:
        assert_almost_equal(data['adjusted_rand'], 0.5714, decimal=4)
    if 'adjusted_mutual_info' in metrics:
        assert_almost_equal(data['adjusted_mutual_info'], 0.4)
    if 'v_measure' in metrics:
        assert_almost_equal(data['v_measure'], 0.8)


@pytest.mark.parametrize('metrics',
                         itertools.combinations(['ratio_duplicates', 'f1_same_duplicates', 'mean_duplicates_count'], 2))
def test_dupdetection_metrics(app, metrics):
    metrics = list(metrics)
    url = V01 + '/metrics/duplicate-detection'
    labels_true = [0, 1, 1, 2, 3, 2]
    labels_pred = [0, 1, 3, 2, 5, 2]

    pars = {'labels_true': labels_true, 'labels_pred': labels_pred, 'metrics': metrics}
    data = app.post_check(url, json=pars)
    assert sorted(data.keys()) == sorted(metrics)
    if 'ratio_duplicates' in metrics:
        assert_almost_equal(data['ratio_duplicates'], 0.5)
    if 'f1_same_duplicates' in metrics:
        assert_almost_equal(data['f1_same_duplicates'], 0.667, decimal=3)
    if 'mean_duplicates_count' in metrics:
        assert_almost_equal(data['mean_duplicates_count'], 0.75)



# -*- coding: utf-8 -*-


def _is_in_range(valid_values):
    """
    Validate that an element is within a given value
    """

    def f(x):
        if x not in valid_values:
            raise ValueError('{} not in {}'.format(x, valid_values))



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import os
import platform
import random

import numpy as np
import pandas as pd
import pickle

from freediscovery.externals.keras_data_utils import _get_file
from freediscovery.externals.keras_data_utils import INTERNAL_DATA_DIR


def _normalize_cachedir(cache_dir):
    """ Normalize the cachedir path. This ensures that the cache_dir
    ends with "ediscovery_cache"
    """
    cache_dir = os.path.normpath(str(cache_dir))
    if 'ediscovery_cache' not in cache_dir:  # not very pretty
        cache_dir = os.path.join(cache_dir, "ediscovery_cache")
    return cache_dir


IR_DATASETS = {'treclegal09_2k_subset': {
                         'md5': '8090cc55ac18fe5c4d5d53d82fc767a2',
                         'size': 2.8},
               'treclegal09_20k_subset': {
                         'md5': '43a711897ce724e873bdbc47a374a57e',
                         'size': 30},
               'treclegal09_37k_subset': {
                         'md5': '9fb6b7505871bbaee5a438de3b0f497c',
                         'size': 55},
               'legal09int': {'md5': '929a675b981282c01c7212030323789f',
                              'size': 1500,
                              'url': "http://r0h.eu/d/legal09int.tar.gz"},
               'fedora_ml_3k_subset': {
                          'md5': '09dbb03d13b8e341bd615ce43f2d836b',
                          'size': 3},
               '20_newsgroups_3categories': {
                          'md5': '7e59e10cbd824190f3f1fa82285c7865',
                          'size': 3,
                          'url': os.path.join(INTERNAL_DATA_DIR, '20_newsgroups_3categories.pkl.xz')  # noqa
                                             },
               '20_newsgroups_micro': {
                          'md5': 'f6ec5e8669ebde1efa11148096c7cc0c',
                          'size': 3,
                          'url': os.path.join(INTERNAL_DATA_DIR, '20_newsgroups_micro.pkl')  # noqa
                                       },
               '20_newsgroups': {
                          'md5': 'f6ec5e8669ebde1efa11148096c7cc0c',
                          'size': 3,
                          'url': 'http://qwone.com/~jason/20Newsgroups/20news-19997.tar.gz'},  # noqa
               }


for name, row in IR_DATASETS.items():
    if 'url' not in row:
        row['url'] = 'https://github.com/FreeDiscovery/FreeDiscovery/' + \
                     'releases/download/v1.1.0/{name}.tar.gz'.format(name=name)


def _load_erdm_ground_truth(outdir):
    """A helper function to load Legal TREC 2009 data"""
    with open(os.path.join(outdir, 'seed_relevant.txt'), 'rt') as fh:
        relevant_files = [el.strip() for el in fh.readlines()]

    with open(os.path.join(outdir, 'seed_non_relevant.txt'), 'rt') as fh:
        non_relevant_files = [el.strip() for el in fh.readlines()]

    if platform.system() == 'Windows':
        relevant_files = [el.replace('/', '\\') for el in relevant_files]
        non_relevant_files = [el.replace('/', '\\')
                              for el in non_relevant_files]
    return non_relevant_files, relevant_files


def _compute_document_id(internal_id, mode):
    if mode == 'squared':
        return internal_id**2
    else:
        raise NotImplementedError


def filter_dict(d, valid_keys):
    return [{key: row[key] for key in row if key in valid_keys} for row in d]


def load_dataset(name='20_newsgroups_3categories', cache_dir='/tmp',
                 verbose=False, verify_checksum=False,
                 document_id_generation='squared', categories=None
                 ):
    """Download a benchmark dataset.

    The currently supported datasets are listed below,

    1. TREC 2009 legal collection

       - `treclegal09_2k_subset`  :   2 400 documents,   2 MB
       - `treclegal09_20k_subset` :  20 000 documents,  30 MB
       - `treclegal09_37k_subset` :  37 000 documents,  55 MB
       - `treclegal09`            : 700 000 documents, 1.2 GB

       The ground truth files for categorization are adapted from TAR Toolkit.

    2. Fedora mailing list (2009-2009)
       - `fedora_ml_3k_subset`

    3. The 20 newsgoups dataset
       - `20_newsgroups_3categories`: only the ['comp.graphics',
       'rec.sport.baseball', 'sci.space'] categories

    If you encounter any issues for downloads with this function,
    you can also manually download and extract the required dataset to
    ``cache_dir`` (the download url is ``http://r0h.eu/d/<name>.tar.gz``),
    then re-run this function to get the required metadata.

    Parameters
    ----------
    name : str, default='20_newsgroups_3categories'
       the name of the dataset file to load
    cache_dir : str, default='/tmp/'
       root directory where to save the download
    verbose : bool, default=False
       print download progress
    verify_checksum : bool, default=False
       verify the checksum of the downloaded archive
    document_id_generation : str
       specifies how the document_id is computed from internal_id
       must be one of ``['identity', 'squared']``
       ``default="identity"`` (i.e. ``document_id = internal_id``)
    categories : str
       select a subsection of the dataset, ``default='all'``

    Returns
    -------

    metadata: dict
       a dictionary containing metadata corresponding to the dataset
    training_set : {dict, None}
       a list of dictionaries for the training set
    test_set : dict
       a list of dictionaries for the test set
    """
    from freediscovery.engine.ingestion import DocumentIndex
    from freediscovery.io import parse_ground_truth_file

    if name not in IR_DATASETS:
        raise ValueError('Dataset name {} not known!'.format(name))

    valid_fields = ['document_id', 'internal_id', 'file_path', 'category']

    has_categories = '20_newsgroups_' in name or 'treclegal09' in name

    # make sure we don't have "ediscovery_cache" in the path
    cache_dir = _normalize_cachedir(cache_dir)
    cache_dir = os.path.dirname(cache_dir)

    outdir = os.path.join(cache_dir, name)
    fname = outdir

    db = IR_DATASETS[name]

    if '20_newsgroups_' in name:
        if db['url'].endswith('.pkl.xz'):
            import lzma
            fname = name + '.pkl.xz'
            opener = lzma.open
        else:
            fname = name + '.pkl'
            opener = open

        with opener(os.path.join(INTERNAL_DATA_DIR, fname), 'rb') as fh:
            twenty_news = pickle.load(fh)

    # Download the dataset if it doesn't exist
    if not os.path.exists(outdir):
        if '20_newsgroups_' in name:
            os.mkdir(outdir)
            for idx, doc in enumerate(twenty_news.data):
                with open(os.path.join(outdir, '{:05}.txt'.format(idx)), 'wt') as fh:  # noqa
                    fh.write(doc)
        else:
            outdir = _get_file(str(fname),
                               db['url'],
                               extract=True,
                               cache_dir=str(cache_dir))
            print('Downloaded {} dataset to {}'.format(name, outdir))

    if 'treclegal09' in name or 'fedora_ml' in name:
        data_dir = os.path.join(outdir, 'data')
    else:
        data_dir = outdir
    md = {'data_dir': str(data_dir), 'name': name}

    di = DocumentIndex.from_folder(str(data_dir))
    di._make_relative_paths()

    training_set = None

    if 'treclegal09' in name:
            negative_files, positive_files = _load_erdm_ground_truth(outdir)

            ground_truth_file = os.path.join(outdir, "ground_truth_file.txt")
            gt = parse_ground_truth_file(str(ground_truth_file))

            res = di.search(gt, drop=False)
            di.data['category'] = res.is_relevant
            di.data['category'] = di.data['category'].apply(
                            lambda x: 'positive' if x == 1 else 'negative')
            di.data['is_train'] = False
            res = di.search(pd.DataFrame({'file_path':
                                          positive_files + negative_files}))
            di.data.loc[res.internal_id.values, 'is_train'] = True
    elif '20_newsgroups_' in name:
        di.data['category'] = np.array(twenty_news.target_names)[twenty_news.target]  # noqa
        di.data['is_train'] = ['-train' in el for el in twenty_news.filenames]

    if categories is not None and has_categories:
        mask = di.data.category.isin(categories)
        di.data = di.data[mask]
        di.data['internal_id'] = np.arange(len(di.data['internal_id']))

    di.data.set_index('internal_id', drop=False, inplace=True)

    di.data['document_id'] = _compute_document_id(di.data['internal_id'],
                                                  document_id_generation)
    di.data = di.data.astype('object')

    if has_categories:
        mask = di.data['is_train']
        training_set = di.render_dict(di.data[mask], return_file_path=True)
        training_set = filter_dict(training_set, valid_fields)
        if name == '20_newsgroups_3categories':
            # make a smaller training set
            random.seed(999998)
            training_set = random.sample(training_set,
                                         min(len(training_set),
                                             di.data.shape[0] // 5))

    dataset = di.render_dict(return_file_path=True)

    dataset = filter_dict(dataset, valid_fields)

    return md, training_set, dataset



# -*- coding: utf-8 -*-

import numpy as np
from sklearn.externals import joblib
import scipy.sparse
import pandas as pd

from freediscovery.engine.base import _BaseWrapper
from freediscovery.utils import setup_model
from freediscovery.cluster import Birch, birch_hierarchy_wrapper
from freediscovery.cluster.utils import _dbscan_noisy2unique
from freediscovery.cluster.utils import centroid_similarity
from freediscovery.cluster.base import _BirchDummy, ClusterLabels


# Clustering methods for FreeDiscovery
# This is highly inspired from the scikit-learn text clustering example

MAX_N_TOP_WORDS = 1000


class _BirchHierarchy(object):
    def __init__(self, model, metric='cosine'):
        self.model = model
        self.htree, _n_clusters = birch_hierarchy_wrapper(model,
                                                          validate=True)
        self._n_clusters = _n_clusters
        self.metric_ = metric

    def fit(self, X):
        """ Compute all the required parameters """
        for row in self.htree.flatten():
            inertia, S_sim = centroid_similarity(X,
                                                 row['document_id_accumulated'],
                                                 nn_metric=self.metric_)
            row['document_similarity'] = S_sim
            row['cluster_similarity'] = inertia


class _BaseClusteringWrapper(object):

    def _merge_response(self, cluster_id):
        res_scores = pd.DataFrame({'internal_id': np.arange(self.fe.n_samples_,
                                                            dtype='int'),
                                   'cluster_id': cluster_id})

        res_scores = res_scores.set_index('internal_id', verify_integrity=True)
        fdb = self.fe.db_.data.set_index('internal_id', verify_integrity=True)

        y = fdb.merge(res_scores,
                      how='inner',
                      left_index=True,
                      right_index=True,
                      suffixes=('_db', '_cluster'))
        return y


class _ClusteringWrapper(_BaseWrapper, _BaseClusteringWrapper):
    """Document clustering

    The algorithms are adapted from scikit learn.

    The option `use_hashing=False` must be set for the feature extraction.
    Recommended options also include, `use_idf=1, sublinear_tf=0, binary=0`.

    Parameters
    ----------
    cache_dir : str
       directory where to save temporary and regression files
    parent_id : str, optional
       dataset id
    mid : str, optional
       model id
    """

    _wrapper_type = "cluster"

    def __init__(self, cache_dir='/tmp/', parent_id=None, mid=None,
                 metric='cosine'):

        super(_ClusteringWrapper, self).__init__(cache_dir=cache_dir,
                                                 parent_id=parent_id,
                                                 mid=mid, load_model=True)

        if self.fe.pars_['use_hashing']:
            raise NotImplementedError('Using clustering with hashed features '
                                      'is not supported by FreeDiscovery!')

        if mid is None:
            self.metric = metric
        else:
            self.metric = self._pars['metric']
        self.km = self.cmod
        del self.cmod
        self._fit_X = None

    def _load_htree(self):
        return joblib.load(str(self.model_dir / self.mid / 'htree'))

    def _cluster_func(self, n_clusters, km, pars=None):
        """ A helper function for clustering, includes base method used by
        all clustering implementations """
        import warnings
        from sklearn.neighbors import NearestCentroid

        new_pars = km.get_params(deep=True)
        new_pars.pop('metric', None)  # dbscan always uses the euclidean metric
        pars.update(new_pars)
        if self._fit_X is None:
            self._fit_X = X = self.pipeline.data
        else:
            X = self._fit_X

        mid, mid_dir = setup_model(self.model_dir)

        with warnings.catch_warnings():
            if type(km).__name__ != "DBSCAN":
                warnings.filterwarnings("ignore", category=DeprecationWarning)
            km.fit(X)

        self.mid = mid
        self.mid_dir = mid_dir

        if type(km).__name__ in ['Birch', '_BirchDummy'] and\
                n_clusters is None:
            # hierarcical clustering, centroids are computed at a later time..
            labels_ = None

        else:
            if type(km).__name__ == "DBSCAN":
                labels_ = _dbscan_noisy2unique(km.labels_)
                n_clusters = len(np.unique(labels_))
                km.labels_ = labels_
            else:
                labels_ = km.labels_

            # i.e. model is not MiniBatchKMeans => compute centroids
            if not hasattr(km, 'cluster_centers_'):
                km.cluster_centers_ = NearestCentroid().fit(X,
                                                            labels_).centroids_
        if type(km).__name__ in ['Birch', '_BirchDummy']:
            if pars['n_clusters'] is None:
                hmod = _BirchHierarchy(km, metric=pars['metric'])
                hmod.fit(X)
                htree = hmod.htree
                km = _BirchDummy()
            else:
                del km.root_
                del km.dummy_leaf_
                htree = None
        else:
            htree = None

        pars['n_clusters'] = n_clusters

        joblib.dump(km, str(self.model_dir / mid / 'model'))
        joblib.dump(htree, str(self.model_dir / mid / 'htree'))
        joblib.dump(pars, str(self.model_dir / mid / 'pars'))

        self.km = km
        self.htree = htree
        self._pars = pars

        return labels_

    def _get_htree(self, X=None, metric='cosine'):
        km = self.km
        method_name = type(km).__name__
        if method_name == 'AgglomerativeClustering':
            htree = {'n_leaves': km.n_leaves_,
                     'n_components': km.n_components_,
                     'children': km.children_.tolist()}
        elif method_name in ['Birch', '_BirchDummy']\
                and self._pars['n_clusters'] is None:
            hmod = _BirchHierarchy(km, metric=metric)
            hmod.fit(X)
            htree = hmod.htree
        else:
            htree = {}
        return htree

    def compute_labels(self, label_method='centroid-frequency', n_top_words=6,
                       cluster_indices=None):
        """ Compute the cluster labels

        Parameters
        ----------
        label_method : str, default='centroid-frequency'
            the method used for computing the cluster labels
        n_top_words : int, default=10
           keep only most relevant n_top_words words
        cluster_indices : list of lists, default=None
           if not None, ignore clustering given by the clustering model
           and compute terms for the cluster provided by the given indices

        Returns
        -------
        cluster_labels : array [n_samples]
        """
        vect = self.fe.vect_

        if 'lsi' in self.pipeline:
            lsi = joblib.load(str(
                              self.pipeline.get_path(self.pipeline['lsi']) /
                              'model'))
        else:
            lsi = None

        lb = ClusterLabels(vect, self.km, lsi=lsi,
                           method=label_method, n_top_words=n_top_words)

        if cluster_indices is not None:
            if self._fit_X is None:
                X = self.pipeline.data
            else:
                X = self._fit_X
            centroids = []
            for indices in cluster_indices:
                X_sl = X[indices]
                centroids.append(X_sl.mean(axis=0))
            if lsi is None:
                centroids = scipy.sparse.csr_matrix(centroids)
            else:
                centroids = np.array(centroids)
        else:
            centroids = None

        terms = lb.predict(centroids=centroids)
        return terms

    def k_means(self, n_clusters, batch_size=1000):
        """
        Perform K-mean clustering

        Parameters
        ----------
        n_clusters : int
           number of clusters
        batch_size : int
           the bath size for the MiniBatchKMeans algorithm
        """
        from sklearn.cluster import MiniBatchKMeans
        pars = {"batch_size": batch_size, 'is_hierarchical': False,
                "metric": self.metric}
        km = MiniBatchKMeans(n_clusters=n_clusters, init='k-means++',
                             n_init=10,
                             init_size=batch_size, batch_size=batch_size)
        return self._cluster_func(n_clusters, km, pars)

    def birch(self, n_clusters=None, threshold=0.5, branching_factor=50,
              max_tree_depth=None):
        """
        Perform Birch clustering

        Parameters
        ----------
        n_clusters : int
            number of clusters
        lsi_components : int
            apply LSA before the clustering algorithm
        threshold : float
            birch threshold
        max_tree_depth : {int, None}
            maximum depth of the hierarchical tree
        """
        pars = {'threshold': threshold, 'is_hierarchical': n_clusters is None,
                'max_tree_depth': max_tree_depth, "metric": self.metric}
        if 'lsi' not in self.pipeline:
            raise ValueError("you must use lsi with birch clustering "
                             "for scaling reasons.")

        if n_clusters is None:
            compute_labels = False
        else:
            compute_labels = True

        km = Birch(n_clusters=n_clusters, threshold=threshold,
                   branching_factor=branching_factor,
                   compute_labels=compute_labels,
                   compute_sample_indices=True)

        return self._cluster_func(n_clusters, km, pars)

    def ward_hc(self, n_clusters, n_neighbors=10):
        """
        Perform Ward hierarchical clustering

        Parameters
        ----------
        n_clusters : int
            number of clusters
        lsi_components : int
            apply LSA before the clustering algorithm
        n_neighbors : int
            N nearest neighbors used for computing the connectivity matrix
        """
        from sklearn.cluster import AgglomerativeClustering
        from sklearn.neighbors import kneighbors_graph
        pars = {'n_neighbors': n_neighbors, 'is_hierarchical': True,
                "metric": self.metric}
        if 'lsi' not in self.pipeline:
            raise ValueError("you must use lsi with birch clustering "
                             "for scaling reasons.")

        # This is really not efficient as
        # it's done a second time in _cluster_func
        X = self.pipeline.data
        connectivity = kneighbors_graph(X, n_neighbors=n_neighbors,
                                        include_self=False)

        km = AgglomerativeClustering(n_clusters=n_clusters, linkage='ward',
                                     connectivity=connectivity)

        return self._cluster_func(n_clusters, km, pars)

    def dbscan(self, n_clusters=None, eps=0.5, min_samples=10,
               algorithm='auto', leaf_size=30):
        """
        Perform DBSCAN clustering

        This can also be used for Duplicate Detection (when ep

        Parameters
        ----------
        n_clusters : int
            number of clusters # not used just present for compatibility
        lsi_components : int
            apply LSA before the clustering algorithm
        eps : float
            The maximum distance between two samples for them to be considered
             as in the same neighborhood.
        min_samples : int
            The number of samples (or total weight) in a neighborhood
            for a point to be considered as a core point.
            This includes the point itself.
        """
        from sklearn.cluster import DBSCAN
        pars = {'is_hierarchical': False, "metric": self.metric}

        km = DBSCAN(eps=eps, min_samples=min_samples, algorithm=algorithm,
                    leaf_size=leaf_size)

        return self._cluster_func(n_clusters, km, pars)

    def scores(self, ref_labels, labels):
        """
        Parameters
        ----------
        ref_labels : list,
            reference labels
        labels : list,
            computed labels
        """
        from sklearn.metrics import (v_measure_score, adjusted_rand_score)
        out = {}
        out['adjusted_rand_score'] = adjusted_rand_score(ref_labels, labels)
        out['v_measure_score'] = v_measure_score(ref_labels, labels)
        return out



import pytest

from freediscovery.engine.utils import validate_mid
from freediscovery.exceptions import WrongParameter


def test_validate_mid():

    validate_mid('test')
    validate_mid('08UIb00')
    with pytest.raises(WrongParameter):
        validate_mid('çd$^a')
    with pytest.raises(WrongParameter):
        validate_mid(' 9023bd')



# -*- coding: utf-8 -*-

import os.path
from unittest import SkipTest

import numpy as np
from numpy.testing import assert_allclose

import pytest
import itertools

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.categorization import _CategorizerWrapper
from freediscovery.engine.near_duplicates import _DuplicateDetectionWrapper
from freediscovery.engine.cluster import _ClusteringWrapper
from freediscovery.engine.lsi import _LSIWrapper
from freediscovery.io import parse_ground_truth_file
from freediscovery.metrics import categorization_score
from freediscovery.exceptions import OptionalDependencyMissing
from freediscovery.tests.run_suite import check_cache


basename = os.path.dirname(__file__)
data_dir = os.path.join(basename, "..", "..", "data", "ds_001", "raw")


@pytest.mark.parametrize('use_hashing, use_lsi, method',
                         itertools.product([False, True],
                                           [False, True],
                                           ['Categorization',
                                            'DuplicateDetection',
                                            'Clustering']))
def test_features_hashing(use_hashing, use_lsi, method):
    # check that models work both with and without hashing

    cache_dir = check_cache()

    n_features = 20000

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(n_features=n_features, use_hashing=use_hashing)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    ground_truth = parse_ground_truth_file(os.path.join(data_dir,
                                           "..", "ground_truth_file.txt"))

    lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=uuid, mode='w')
    lsi_res, exp_var = lsi.fit_transform(n_components=100)
    assert lsi._load_pars() is not None
    lsi._load_model()

    if method == 'Categorization':
        if use_lsi:
            parent_id = lsi.mid
            method = 'NearestNeighbor'
        else:
            parent_id = uuid
            method = 'LogisticRegression'
        cat = _CategorizerWrapper(cache_dir=cache_dir, parent_id=parent_id,
                                  cv_n_folds=2)
        cat.fe.db_.filenames_ = cat.fe.filenames_
        index = cat.fe.db_._search_filenames(ground_truth.file_path.values)

        try:
            coefs, Y_train = cat.fit(
                                    index,
                                    ground_truth.is_relevant.values,
                                    method=method
                                    )
        except OptionalDependencyMissing:
            raise SkipTest

        Y_pred, md = cat.predict()
        X_pred = np.arange(cat.fe.n_samples_, dtype='int')
        idx_gt = cat.fe.db_._search_filenames(ground_truth.file_path.values)

        scores = categorization_score(idx_gt,
                                      ground_truth.is_relevant.values,
                                      X_pred, np.argmax(Y_pred, axis=1))
        assert_allclose(scores['precision'], 1, rtol=0.5)
        assert_allclose(scores['recall'], 1, rtol=0.7)
        cat.delete()
    elif method == 'DuplicateDetection':
        dd = _DuplicateDetectionWrapper(cache_dir=cache_dir, parent_id=uuid)
        try:
            dd.fit()
        except ImportError:
            raise SkipTest
        cluster_id = dd.query(distance=10)
    elif method == 'Clustering':
        if not use_hashing:
            if use_lsi:
                parent_id = lsi.mid
                method = 'birch'
            else:
                parent_id = uuid
                method = 'k_means'
            cat = _ClusteringWrapper(cache_dir=cache_dir, parent_id=parent_id)
            cm = getattr(cat, method)
            labels = cm(2)

            htree = cat._get_htree(cat.pipeline.data)

            terms = cat.compute_labels(n_top_words=10)
        else:
            with pytest.raises(NotImplementedError):
                _ClusteringWrapper(cache_dir=cache_dir, parent_id=uuid)
    else:
        raise ValueError




from itertools import product
from unittest import SkipTest

import pytest
import numpy as np
from numpy.testing import assert_allclose
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.utils.sparsefuncs_fast import csr_row_norms

from freediscovery.feature_weighting import _smart_tfidf, SmartTfidfTransformer
from freediscovery.feature_weighting import _validate_smart_notation

documents = ["Shipment of gold damaged in aa fire.",
             "Delivery of silver arrived in aa silver truck.",
             "Shipment of gold arrived in aa truck.",
             "short sentence"
             ]


# all division by zeros should be explicitly handled
@pytest.mark.parametrize('scheme, array_type', product(("".join(el)
                         for el in product('nlabLd', 'ntspd',
                                           ['n', 'c', 'l', 'u',
                                            'cp', 'lp', 'up'])),
                        ['sparse']))
@pytest.mark.filterwarnings('error')
def test_feature_weighting_empty_document(scheme, array_type):
    documents_new = documents + ['']
    tf = CountVectorizer().fit_transform(documents_new)
    # check that all weightings preserve zeros rows (with no tokens)
    # and that no overflow warnings are raised
    X = _smart_tfidf(tf, scheme)
    assert_allclose(X.A[-1], np.zeros(tf.shape[1]))


@pytest.mark.parametrize('weighting', ['nncp', 'nnlp', 'nnup'])
def test_pivoted_normalization(weighting):
    tf = CountVectorizer().fit_transform(documents)
    X_ref = _smart_tfidf(tf, 'nnc')
    if weighting == 'nncp':
        # pivoted cosine normalization == cosine normalization
        # when alpha == 1.0
        X = _smart_tfidf(tf, 'nncp', norm_alpha=1.0)
        assert_allclose(X.A, X_ref.A)
    X = _smart_tfidf(tf, weighting, norm_alpha=0.75)
    # shorter documents (last one) gets normalized by a larger value
    assert (X[-1].data < X_ref[-1].data).all()


def test_smart_tfidf_transformer_compatibility():
    raise SkipTest

    try:
        from sklearn.utils.estimator_checks import check_estimator
    except ImportError:
        raise SkipTest
    check_estimator(SmartTfidfTransformer)


@pytest.mark.parametrize('scheme', ("".join(el)
                                    for el in product('nlabLd', 'ntspd',
                                                      ['n', 'c', 'l', 'u',
                                                       'cp', 'lp', 'up']))
                         )
def test_smart_tfidf_transformer(scheme):
    tf = CountVectorizer().fit_transform(documents)

    estimator = SmartTfidfTransformer(weighting=scheme)

    X = estimator.fit_transform(tf)

    scheme_t, scheme_d, scheme_n = _validate_smart_notation(scheme)
    if scheme_d not in 'dp':
        # the resulting document term matrix should be positive
        # (unless we use probabilistic idf weighting)
        assert (X.A >= 0).all()

    # norm cannot be zero
    X_norm = csr_row_norms(X)
    assert (X_norm > 0).all()

    X_ref = None
    if scheme == 'nnn':
        X_ref = X
    elif scheme == 'nnc':
        X_ref = TfidfVectorizer(use_idf=False, smooth_idf=False).fit_transform(documents)
    elif scheme == 'ntc':
        X_ref = TfidfVectorizer(use_idf=True, smooth_idf=False).fit_transform(documents)
    elif scheme == 'lnn':
        X_ref = TfidfVectorizer(use_idf=False, sublinear_tf=True,
                                smooth_idf=False, norm=None).fit_transform(documents)
    elif scheme == 'ltc':
        X_ref = TfidfVectorizer(use_idf=True, sublinear_tf=True,
                                smooth_idf=False).fit_transform(documents)
    elif scheme == 'ltl':
        X_ref = TfidfVectorizer(use_idf=True, sublinear_tf=True,
                                smooth_idf=False, norm='l1').fit_transform(documents)

    if X_ref is not None:
        assert_allclose(X.A, X_ref.A, rtol=1e-7, atol=1e-6)

    assert len(estimator.dl_) == tf.shape[0]
    assert len(estimator.du_) == tf.shape[0]
    if scheme_d in ['tsp']:
        assert len(estimator.df_) == tf.shape[1]

    X_2 = SmartTfidfTransformer(weighting=scheme).fit(tf).transform(tf)
    assert_allclose(X.A, X_2.A, rtol=1e-6, atol=1e-6)

    if scheme_d in 'stp':
        assert estimator.df_ is not None

    sl = slice(2)
    tf_w_sl = estimator.transform(tf[sl])
    assert_allclose(X[sl].A, tf_w_sl.A)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from unittest import SkipTest


def test_get_positions():
    from freediscovery.interpretation import _get_keyword_positions

    textline = "Hello world, this is Eric."
    keywords = ["world", "eric"]
    positions = _get_keyword_positions(textline, keywords)
    assert(positions == [(6, 11), (21, 25)])

    textline = "email me at eric_bass@some-email.com"
    keywords = ["eric", "bass", "eric_bass"]
    positions = _get_keyword_positions(textline, keywords)
    assert(positions == [(12, 16), (12, 21), (17, 21)])

    textline = "administrative procedures"
    keywords = ["administrative", "mini", "procedure", "pro"]
    positions = _get_keyword_positions(textline, keywords)
    assert(positions == [(0, 14), (2, 6), (15, 18), (15, 24)])


def test_overlap_deduplication():
    from freediscovery.interpretation import _keep_longest_overlapping_substrings

    positions = [(6, 11), (21, 25)]
    deduplicated_positions = _keep_longest_overlapping_substrings(positions)
    assert(deduplicated_positions == [(6, 11), (21, 25)])

    positions = [(12, 16), (12, 21), (17, 21)]
    deduplicated_positions = _keep_longest_overlapping_substrings(positions)
    assert(deduplicated_positions == [(12, 21)])

    positions = [(0, 14), (2, 6), (15, 18), (15, 24)]
    deduplicated_positions = _keep_longest_overlapping_substrings(positions)
    assert(deduplicated_positions == [(0, 14), (15, 24)])


def test_explain_categorization():
    from freediscovery.interpretation import explain_categorization, _create_random_weights

    document_text = """
    Hello world\nthis is test\n
    From:           Eric Bass </o=teneo-test/ou=first administrative group/cn=recipients/cn=bass_eric>
    Sent:           Wed Jul 12 2000 15:30:00 GMT
    CC:             Subject:        Draft Order Determined
    """
    words_weights = _create_random_weights(document_text, 0.2)

    def colormap_mock(x):
       return (1.0, 1.0, 1.0, 1.0)

    try:
        document_html = explain_categorization(words_weights, document_text, colormap=colormap_mock)
    except ImportError:
        raise SkipTest
    assert len(document_html) > 0



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import os
import sys
import shutil
from contextlib import contextmanager
import numpy as np
import uuid
from pkg_resources import parse_version
import sklearn
from .exceptions import WrongParameter

# this is wrong, and should be eventually replaced
INT_NAN = -99999

sklearn_version = parse_version(sklearn.__version__)


@contextmanager
def _silent(stream='stderr'):
    stderr = getattr(sys, stream)
    fh = open(os.devnull, 'w')
    sys.stderr = fh
    yield
    setattr(sys, stream, stderr)


def _rename_main_thread():
    """
    This aims to address the fact that joblib wrongly detects uWSGI workers
    as running in the non main thread even when they are not
    see https://github.com/joblib/joblib/issues/180
    """
    import threading
    if isinstance(threading.current_thread(), threading._MainThread) and \
       threading.current_thread().name != 'MainThread':
        print('Warning: joblib: renaming current thread {} to "MainThread".'
              .format(threading.current_thread().name))
        threading.current_thread().name = 'MainThread'


def _count_duplicates(x):
    """Return y an array of the same shape as x with the number of
    duplicates for each element"""
    _, indices, counts = np.unique(x, return_counts=True, return_inverse=True)
    return counts[indices]


def generate_uuid(size=16):
    """
    Generate a unique id for the model
    """
    sl = slice(size)
    return uuid.uuid4().hex[sl]  # a new random id


def setup_model(base_path, mid=None, mode='w'):
    """
    Generate a unique model id and create the corresponding folder
    for storing results
    """
    if mode not in ['w', 'fw']:
        raise WrongParameter(('Cannot setup model {} in mode={} '
                              'must be one of "w", "fw"!')
                             .format(base_path, mode))
    if mid is None:
        mid = generate_uuid()

    if not base_path.exists():
        base_path.mkdir()

    mid_dir = base_path / mid
    if mid_dir.exists():
        if mode == 'w':
            raise WrongParameter(('model mid={} already exists in {} .'
                                  'Use mode="fw" to overwrite!')
                                 .format(mid, base_path))
        shutil.rmtree(str(mid_dir))  # removing the old folder nevertheless
    mid_dir.mkdir()
    return mid, mid_dir


def _docstring_description(docstring):
    """ Given a function docstring, return only the text prior
    to the "Parameters" section"""

    res = []
    for line in docstring.splitlines():
        if line.strip() == 'Parameters':
            break
        res.append(line)
    return '\n'.join(res)


def _query_features(vect, X, indices, n_top_words=10, remove_stop_words=False):
    """ Query the features with most weight

    Parameters
    ----------
    vect : TfidfVectorizer
       the vectorizer object
    X : ndarray
       the document term tfidf array
    indices : list or ndarray
      indices for the subcluster
    n_top_words : int
      the number of workds to return
    remove_stop_words : bool
      remove stop words
    """
    from .cluster.base import select_top_words

    # this should raise a warning when used with wrong weights
    X = X[indices]

    centroid = X.sum(axis=0).view(type=np.ndarray)[0] / len(indices)
    order_centroid = centroid.argsort()[::-1]
    terms = vect.get_feature_names()

    out = []
    for ridx, idx in enumerate(order_centroid):
        if len(out) >= n_top_words:
            break
        if remove_stop_words:
            out += select_top_words([terms[idx]])
        else:
            out.append(terms[idx])
    return out


def _type_mapper(mtype):
    mapper = {'unicode': 'str', 'long': 'str'}
    if mtype in mapper:
        return mapper[mtype]
    else:
        return mtype


def dict2type(d, collapse_lists=False, max_depth=10):
    """Recursively walk though the object
    and replace all dict values by their type

    Parameters
    ----------
    collapse_lists : bool
      collapse a list to a single element
    max_depth : bool
      maximum depth on which the typing would be computed
    """
    if max_depth == 0:
        return _type_mapper(type(d).__name__)

    if isinstance(d, dict):
        res = {}
        for key, val in d.items():
            res[key] = dict2type(val, collapse_lists, max_depth - 1)
        return res
    elif isinstance(d, list):
        res = [dict2type(el, collapse_lists, max_depth - 1) for el in d]
        if collapse_lists:
            res = list(set(res))
        return res
    else:
        return _type_mapper(type(d).__name__)


def sdict_keys(x):
    """Sorted dictionary keys of x"""
    return list(sorted(x.keys()))


def _mean_csr_nonzero_axis1(X):
    """ Compute the mean of non zero elements a csr array along axis=1
    """

    sums = X.sum(axis=1).A1
    counts = np.diff(X.indptr)
    # prevent 0 / 0 division (which should yield 0 here)
    counts[counts == 0] = 1
    return sums / counts


def _paginate(df, batch_id, batch_size):
    """ Given a dataframe df, return a batch of rows
    specified by `batch_id` and `batch_size`

    Parameters
    ----------
    df : pd.DataFrame
       the input dataframe
    batch_id : int
       the batch_id element
    batch_size : int
       the batch size

    Returns
    -------
    df_out : pd.DataFrame
       sliced dataframe
    md : dict
       a dictionationary with pagination information
    """
    n_samples = df.shape[0]
    if batch_id >= 0:
        mslice = slice(batch_id*batch_size,
                       min((batch_id + 1)*batch_size, n_samples))
        df_out = df.iloc[mslice, :]
    else:
        df_out = df

    pagination = {'batch_id': batch_id,
                  'current_response_count': df_out.shape[0],
                  'total_response_count': n_samples}
    if batch_id < 0:
        pagination['batch_id_last'] = batch_id
    else:
        pagination['batch_id_last'] = n_samples // batch_size

    return df_out, pagination


def _split_path(path):
    """ A helper function that splits the path into a list

    Parameters
    ----------
    path : str
      path to split
    """
    path = os.path.normpath(path)

    head, tail = os.path.split(path)

    if not head:
        return [tail]
    elif head == path:
        if tail:
            return [head, tail]
        else:
            return [head]
    else:
        if not tail:
            return _split_path(head)
        else:
            return _split_path(head) + [tail]



# -*- coding: utf-8 -*-


def binary_sensitivity_analysis(estimator, vocabulary, X_row):
    """Explain the binary categorization results

    Parameters
    ----------
    estimator : sklearn.base.BaseEstimator
      the binary categorization estimator
      (must have a `decision_function` method)
    vocabulary : list [n_features]
      vocabulary (list of words or n-grams)
    X_row : sparse CSR ndarray [n_features]
      a row of the document term matrix
    """
    if X_row.ndim != 2 or X_row.shape[0] != 1:
        raise ValueError('X_row must be an 2D sparse array,'
                         'with shape (1, N) not {}'.format(X_row.shape))
    if X_row.shape[1] != len(vocabulary):
        raise ValueError(('The vocabulary length ({}) does not match '
                          'the number of features in X_row ({})')
                         .format(len(vocabulary), X_row.shape[1]))

    vocabulary_inv = {ind: key for key, ind in vocabulary.items()}

    if type(estimator).__name__ == 'LogisticRegression':
        coef_ = estimator.coef_
        if X_row.shape[1] != coef_.shape[1]:
            raise ValueError(("Coefficients size {} does "
                              "not match n_features={}")
                             .format(coef_.shape[1], X_row.shape[1]))

        indices = X_row.indices
        weights = X_row.data*estimator.coef_[0, indices]
        weights_dict = {}
        for ind, value in zip(indices, weights):
            key = vocabulary_inv[ind]
            weights_dict[key] = value
        return weights_dict
    else:
        raise NotImplementedError()



"""
.. _optimize_tfidf_scheme_example:

Optimizing TF-IDF schemes
=========================

An example of optimizing TF-IDF weighting schemes using
5 fold cross-validation
"""
from __future__ import print_function

import os
from itertools import product

import numpy as np
from sklearn.datasets import fetch_20newsgroups
from sklearn.linear_model import LogisticRegression
from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer
from sklearn.model_selection import train_test_split, GridSearchCV
from sklearn.pipeline import Pipeline

from freediscovery.feature_weighting import SmartTfidfTransformer

rng = np.random.RandomState(34)

###############################################################################
#
# We load and vectorize 2 classes from the 20 newsgroup dataset,

newsgroups = fetch_20newsgroups(subset='train',
                                categories=['sci.space', 'comp.graphics'])
vectorizer = CountVectorizer(stop_words='english')
X = vectorizer.fit_transform(newsgroups.data)

###############################################################################
#
# then compute baseline categorization performance using Logistic Regression
# and the TF-IDF transfomer from scikit-learn

X_tfidf = TfidfTransformer().fit_transform(X)

X_train, X_test, y_train, y_test = train_test_split(X, newsgroups.target,
                                                    random_state=rng)

pipe = Pipeline(steps=[('tfidf', TfidfTransformer()),
                       ('logisticregression', LogisticRegression())])

pipe.fit(X_train, y_train)
print('Baseline TF-IDF categorization accuracy: {:.3f}'
      .format(pipe.score(X_test, y_test)))

###############################################################################
#
# Next, we search, using 5 fold cross-validation, for the best TF-IDF weighting
# scheme among the 80+ combinations supported by
# :class:`~freediscovery.feature_weighting.SmartTfidfTransformer`. Two
# hyper-parameters are worth optimizing in this case,
#
# * ``weighting`` parameter that defines the TF-IDF weighting (see the
#   :ref:`tfidf_section` user manual section for more details)
# * ``norm_alpha`` is the α parameter in the pivoted normalization
#   when ``weighting=="???p"``.
#
# To reduce the search parameter space in this example, we also can exclude
# the case when either the term weighting, feature weighing or normalization is
# not used as it expected to yield worse than baseline performance. We also
# exclude the non smoothed IDF weightings (``?t?``, ``?p?``) since thay return
# NaNs when some of the document frequency is 0 (which will be the case
# during cross-validation). Finally, by noticing
# that the case ``xxxp`` with  ``norm_alpha=1.0`` corresponds to the weighing
# ``xxx`` (i.e. with pivoted normalization disabled) we can reduce the search
# space even further.

pipe = Pipeline(steps=[('tfidf', SmartTfidfTransformer()),
                       ('logisticregression', LogisticRegression())])

param_grid = {'tfidf__weighting': ["".join(el) + 'p'
                                   for el in product('labLd', 'sd',
                                                     "clu")],
              'tfidf__norm_alpha': np.linspace(0, 1, 10)}

pipe_cv = GridSearchCV(pipe,
                       param_grid=param_grid,
                       verbose=1,
                       n_jobs=(1 if os.name == 'nt' else -1),
                       cv=5)
pipe_cv.fit(X_train, y_train)
print('Best CV params: weighting={weighting}, norm_alpha={norm_alpha:.3f} '
      .format(**pipe_cv.best_estimator_.steps[0][1].get_params()))
print('Best TF-IDF categorization accuracy: {:.3f}'
      .format(pipe_cv.score(X_test, y_test)))


###############################################################################
#
# In this example, by tuning TF-IDF weighting scheme with pivoted
# normalization, we obtain a categorization accuracy score of 0.99 as compared
# to a baseline TF-IDF score of 0.973. It is also interesting to notice that
# the best weighting hyper-parameter in this case is ``lnup`` which
# corresponds to the "unique pivoted normalization" case proposed by
# Singhal *et al.* (1996), although with a different α value.



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import numpy as np
from sklearn.utils.validation import check_array


def _binary_linkage2clusters(linkage, n_samples):
    """ Given a list of elements of size n_sample and a linkage matrix
    linking some of those samples, compute the cluster_id of each element

    Parameters
    ----------

    linkage : array (n_pairs, 2)
       arrays indicating binary links between elements
    n_samples : int
       total number of elements

    Returns
    -------
    labels : array (n_samples)
    """

    if (linkage > n_samples).any():
        raise ValueError
    if (linkage < 0).any():
        raise ValueError
    if n_samples < 0:
        raise ValueError

    dmap = {}
    idx = 0
    for a, b in linkage:
        if a in dmap:
            cid = dmap[a]
        elif b in dmap:
            cid = dmap[b]
        else:
            cid = idx
            idx += 1
        dmap[a] = cid
        dmap[b] = cid

    labels = np.zeros(n_samples, dtype=np.int)
    cid = 0
    for idx in range(n_samples):
        if idx in dmap:
            labels[idx] = n_samples + dmap[idx]
        else:
            labels[idx] = cid
            cid += 1
    _, labels_renamed = np.unique(labels, return_inverse=True)
    return labels_renamed


def _merge_clusters(X, rename=False):
    """
    Compute a union of all clusters

    Used to determine which cluster_id a document should belong to
    if at least one of it's lexicons suggest that it's a duplicate

    Approximate time complexity O(n_samples*n_features)

    Parameters
    ----------
     X: array (n_samples, n_features)
       input arrays with the cluster id's to merge
     rename : binary
       make sure the output array is between 0 and len(unique(cluster_id))

    Parameters
    ----------
      cluster_id: array (n_samples)
    """
    X = check_array(X, ensure_2d=True)

    n_samples, n_features = X.shape

    y = np.zeros(n_samples, dtype=X.dtype)

    out = {}

    for (i_idx, X_row) in enumerate(X):
        for j_idx, X_el in enumerate(X_row):
            if (j_idx, X_el) in out:
                res = out[(j_idx, X_el)]
                break
        else:
            res = X_row[0]  # use the 1st columnt index for this cluster id

        for (j_idx, X_el) in enumerate(X_row):
            out[(j_idx, X_el)] = res

        y[i_idx] = res
    if rename:
        _, labels_renamed = np.unique(y, return_inverse=True)
        return labels_renamed
    else:
        return y


def _dbscan_noisy2unique(labels_):
    """
    Take labels_ given by DBSCAN and replace each "noisy"
    point specified by -1, to a unique cluster id
    """
    labels_ = np.asarray(labels_).copy()
    mask = labels_ == -1
    indices = np.arange(mask.sum(), dtype=np.int)
    indices += labels_.max()+1
    labels_[mask] = indices
    return labels_


def _dbscan_unique2noisy(labels_):
    """
    Given an array of cluster_id, for each element that forms a cluster
    by itself, replace the corresponding cluster_id by -1.

    This is the iverse operation to _dbscan_noisy2unique
    """
    from collections import Counter
    labels_ = np.asarray(labels_).copy()
    cnt = Counter()
    for el in labels_:
        cnt[el] += 1

    labels_ndup_ = np.zeros(labels_.shape, dtype=labels_.dtype)
    for idx, el in enumerate(labels_):
        el_cnt = cnt[el]
        if el_cnt > 1:
            labels_ndup_[idx] = el
        elif el_cnt == 1:
            labels_ndup_[idx] = -1
        else:
            raise ValueError('This should not be possible!')
    return labels_ndup_


def centroid_similarity(X, internal_ids, nn_metric='cosine'):
    """ Given a list of documents in a cluster, compute the cluster centroid,
    intertia and individual distances

    Parameters
    ----------
    internal_ids : list
      a list of internal ids
    nn_metric : str
      a rescaling of the metric if needed
    """
    from ..metrics import _scale_cosine_similarity
    from sklearn.metrics.pairwise import pairwise_distances

    X_sl = X[internal_ids, :]
    centroid = X_sl.mean(axis=0)

    if centroid.ndim == 1:
        centroid = centroid[None, :]

    S_cos = 1 - pairwise_distances(X_sl, centroid, metric='cosine')
    S_sim = _scale_cosine_similarity(S_cos, metric=nn_metric)
    S_sim_mean = np.mean(S_sim)
    return float(S_sim_mean), S_sim[:, 0]



# -*- coding: utf-8 -*-

import pytest
from unittest import SkipTest
from numpy.testing import assert_equal, assert_almost_equal

from ...utils import _silent, dict2type, sdict_keys
from .base import parse_res, V01, app, app_notest, get_features_cached, get_features_lsi


#=============================================================================#
#
#                     Exception handling
#
#=============================================================================#


@pytest.mark.parametrize("method", ['feature-extraction', 'categorization', 'lsi', 'clustering'])
def test_get_model_404(app_notest, method):
    method = V01 + "/{}/DOES_NOT_EXISTS".format(method)
    with _silent('stderr'):
        res = app_notest.get(method)

    data = parse_res(res)
    assert res.status_code in [500, 404]  # depends on the url
    #assert '500' in data['message']

    assert sorted(data.keys()) == sorted(['messages'])


@pytest.mark.parametrize("method", ['feature-extraction', 'categorization', 'lsi', 'clustering'])
def test_get_model_train_404(app_notest, method):
    method = V01 + "/{}/DOES_NOT_EXISTS/DOES_NOT_EXIST".format(method)
    with _silent('stderr'):
        res = app_notest.get(method)

    assert res.status_code in [404, 500]


@pytest.mark.parametrize("method", ['feature-extraction', 'categorization', 'lsi', 'clustering'])
def test_get_model_predict_404(app_notest, method):

    method = V01 + "/{0}/DOES_NOT_EXISTS/DOES_NOT_EXIST/predict".format(method)
    with _silent('stderr'):
        res = app_notest.get(method)

    assert res.status_code == 404

    method = V01 + "/{0}/DOES_NOT_EXISTS/DOES_NOT_EXIST/test".format(method)
    with _silent('stderr'):
        res = app_notest.post(method)

    assert res.status_code == 404


def test_exception_handling(app_notest):
    dsid, pars, _ = get_features_cached(app_notest)

    method = V01 + "/categorization/"
    with _silent('stderr'):
        res = app_notest.post(method,
                        json={
                              'parent_id': dsid,
                              'index': [0, 0, 0],       # just something wrong
                              'y': ['ds', 'dsd', 'dsd'],
                              'method': "LogisticRegression",
                              'cv': 0,
                              })
    data = parse_res(res)
    assert res.status_code in [500, 422]
    assert sorted(data.keys()) == ['messages']



# -*- coding: utf-8 -*-

import os
import os.path

from sklearn.externals.joblib import dump, load

from freediscovery.engine.pipeline import PipelineFinder


class _StopWordsWrapper(object):
    """A mechanism for adding / managing custom stop words
        Parameters
        ----------
        cache_dir : str
           folder where the model will be saved
        stop_words : list
           a list of strings
    """
    _wrapper_type = "stop_words"

    def __init__(self, cache_dir='/tmp/'):
        """ Initialize a stop words wrapper

        Parameters
        ----------
        cache_dir : str
          the cache directory
        """
        self.cache_dir = PipelineFinder._normalize_cachedir(cache_dir)
        self.model_dir = self.cache_dir / 'stop_words'

        if not self.model_dir.exists():
            self.model_dir.mkdir()

    def save(self, name, stop_words):
        """
        Save a list of stop_words with joblib.save under
             $CACHE_DIR/stop_words/<name>.pkl

        Parameters
        ----------
        name : str
            stop words name / identifier
        stop_words : list
            list of stop words
        """

        self.stop_words = stop_words  # list of stop words

        self.name = self.model_dir / (str(name) + '.pkl')

        dump(self.stop_words, str(self.name))

    def load(self, name):
        """Retrive stop words specified by a name
        """
        self.name = self.model_dir / (str(name) + '.pkl')
        self.stop_words = load(str(self.name))
        return (self.stop_words)

    def delete(self, name):
        """Delete stop words specified by a name
        """
        if (self.model_dir / (str(name) + '.pkl')).exists():
            (self.model_dir / (str(name) + '.pkl')).unlink()

    def __contains__(self, name):
        """ Check if a given stop words set exists """
        return (self.model_dir / (str(name) + '.pkl')).exists()

    def list(self):
        """ Returns a list of exiting stop-words """
        return [os.splitext(el)[0] for el in
                os.listdir(str(self.model_dir))]



# Authors: Roman Yurchak
#
# License: BSD 3 clause

from itertools import chain
from textwrap import dedent

from .birch import Birch
from freediscovery.externals.jwzthreading import Container
from sklearn.exceptions import NotFittedError


class BirchSubcluster(Container):
    """A container class for BIRCH cluster hierarchy

    This is a dict like container, that is used to store each subcluster in
    the cluster hierarchy computed by :class:`freediscovery.cluster.Birch`. A
    given subcluster links to the parent / children subclusters in the
    hierarchy with the following attributes,

     * `parent` : :class:`BirchSubcluster`, the parent
       container
     * `children` : ``list`` of :class:`BirchSubcluster`,
       contains the children subclusters

    Each subcluster stores the following dictionary keys,

     * `document_id` : ``list``, a list of document / sample ids contained
       in this subcluster (excluding its children).
     * `document_id_accumulated`:  a list of document / sample ids
       contained in this subcluster and its children. Only available when
       this class was build using :func:`birch_hierarchy_wrapper` with
       the ``compute_document_id=True`` parameter.  It can be re-computed
       with the ``document_id_accumulated`` class property.
     * `cluster_size`: int, the number of samples contained in this
       subcluster and its children. This corresponds to the length of
       the ``document_id_accumulated`` property.  Only available when
       this class was build using :func:`birch_hierarchy_wrapper` with
       the ``compute_document_id=True`` parameter.

    other keys may be user-computed as necessary.

    See :ref:`User Manual <exploring_hierarchical_tree_section>` for
    more details.

    Note
    ----
    This class descends from
    :class:`freediscovery.externals.jwzthreading.Container`
    originally used to represent e-mail threads obtained with the JWZ
    algorithm in
    `jwzthreading <https://github.com/FreeDiscovery/jwzthreading>`_,
    though it is general enough to represent other hierarchical
    stuctures, such as BIRCH cluster hierarchy.

    In FreeDiscovery this class is primarily used for documents. As a
    result the variables/methods containing the term "document"
    have the same meaning as "sample" in the general scikit-learn context.
    """

    @property
    def document_count(self):
        """Count of all documents in the children subclusters"""
        partial_sum = sum([child.document_count for child in self.children])
        return len(self.get('document_id', [])) + partial_sum

    @property
    def document_id_accumulated(self):
        """Returns  list of document / sample ids contained
        in this subcluster or any of its children."""
        partial_sum = chain.from_iterable(el.document_id_accumulated
                                          for el in self.children)
        return list(self.get('document_id', [])) + list(partial_sum)

    def limit_depth(self, max_depth=None):
        """ Truncate the tree to the provided maximum depth

        Parameters
        ----------
        max_depth : int
          hierarchy depth to which truncate the tree
        """
        if self.current_depth >= max_depth:
            self.children = []

        for el in self.children:
            el.limit_depth(max_depth)

    def display_tree(self, max_depth=None):
        """Print the content of hierarchical tree below this subcluster
        """
        _print_container(self)

    def __repr__(self):
        content = super(BirchSubcluster, self).__repr__()
        if self.parent:
            parent_repr = 'BirchSubcluster[subcluster_id={}]'\
                    .format(self.parent['cluster_id'])
        else:
            parent_repr = 'None'

        if self.children:
            child_repr = ', '.join([
                  'BirchSubcluster[cluster_id={}]'
                  .format(ctr['cluster_id'])
                  for ctr in self.children])
        else:
            child_repr = ''

        return dedent("""
             {}
               * parent: {}
               * children: [{}]
             """.format(content, parent_repr, child_repr))

    def increment_cluster_id(self, value):
        """ Increment the cluster_id of all children
        by the given value
        """
        self['cluster_id'] += value
        for child in self.children:
            child.increment_cluster_id(value)


def _check_birch_tree_consistency(node):
    """ Check that the _id we added is consistent """
    for el in node.subclusters_:
        if el.samples_id_ is None:
            raise ValueError('Birch was fitted without storing samples. '
                             'Please re-initalize Birch with '
                             'compute_sample_indices=True !')
        if el.n_samples_ != len(el.samples_id_):
            raise ValueError(('For subcluster ',
                             '{}, n_samples={} but len(id_)={}')
                             .format(el, el.n_samples_, el.samples_id_))
        if el.child_ is not None:
            _check_birch_tree_consistency(el.child_)


def _birch_hierarchy_constructor(node, depth=0, cluster_id=0,
                                 container=BirchSubcluster,
                                 prune_single_clusters=True):
    """Wrap BIRCH cluster hierarchy with a container class

    Parameters
    ----------
    node : _CFNode
      a node in the birch hierarchy tree
    depth : int
      cluster depth
    cluster_id : int
      cluster id
    container : freediscovery.cluster.BirchSubcluster, default=BirchSubcluster
      a subclass of :class:`~freediscovery.cluster.BirchSubcluster`
      that will be used to wrap each BIRCH subcluster
    prune_single_clusters : bool, default=True
      if a cluster has a single child cluster, account only the child in the hierarchy
      (not the parent). This removes identical clusters at different depth, however when
      enabled the number of documents at a given depth d is no longer preserved. 

    Returns
    -------
    res : BirchSubcluster
      a hierarchical structure with the resulting clustering
    n_subclusters : int
      the total number of subclusters
    """

    htree = container()
    htree['document_id'] = document_id_list = []
    htree['cluster_id'] = cluster_id
    htree.prune_single_clusters = prune_single_clusters
    # detect if this subcluster has a single child

    for el in node.subclusters_:
        if el.child_ is not None:
            cluster_id += 1
            subtree, cluster_id = _birch_hierarchy_constructor(
                     el.child_, depth=depth+1, cluster_id=cluster_id,
                     container=container,
                     prune_single_clusters=prune_single_clusters)
            if len(subtree.children) == 1 and prune_single_clusters:
                # we are going to skip the single child subcluster,
                # so don't need to increment the cluster_id
                subtree.increment_cluster_id(-1)
                cluster_id += -1
                # skip the single child subcluster
                htree.add_child(subtree.children[0])
            else:
                htree.add_child(subtree)
        else:
            document_id_list += el.samples_id_
    if depth == 0:
        # make sure we return the correct number of clusters
        cluster_id += 1
    return htree, cluster_id


def birch_hierarchy_wrapper(birch, container=BirchSubcluster, validate=True,
                            compute_document_id=True):
    """Wrap BIRCH cluster hierarchy with a container class

    This class as inpu

    Parameters
    ----------
    birch : freediscovery.cluster.Birch
      a trained Birch clustering
    container : freediscovery.cluster.BirchSubcluster, default=BirchSubcluster
      a subclass of :class:`~freediscovery.cluster.BirchSubcluster`
      that will be used to wrap each BIRCH subcluster
    validate : bool, default=True
      check the consistency of the constructed hierarchical tree
      (this may carry some overhead)
    compute_document_id : bool, default=True
      compute the document/sample ids belonging to each subcluster

    Returns
    -------
    htree : BirchSubcluster
      a container containing the data from the BIRCH cluster hierarchy
    n_subclusters : int
      the total number of subclusters

    Note
    ----
    in FreeDiscovery BIRCH if primairly used to cluster documents. As a
    result the variables/methods containing the term "document"
    have the same meaning as "sample" in the general scikit-learn context.
    """
    if not isinstance(birch, Birch):
        raise ValueError('the birch object must be created with '
                         'freediscovery.cluster.Birch')

    if not hasattr(birch, "root_"):
        raise NotFittedError("The Birch model must be fitted first!")

    if validate:
        _check_birch_tree_consistency(birch.root_)

    htree, n_subclusters = _birch_hierarchy_constructor(birch.root_,
                                                        container=container)
    if validate:
        if len(htree.document_id_accumulated) != birch.n_samples_:
            print(htree.document_id_accumulated)
            raise ValueError(("Building hierarchy failed: root node contains "
                              "{} documents, while the total document number "
                              "is {}")
                             .format(len(htree.document_id_accumulated),
                                     birch.n_samples_))
    if compute_document_id:
        for row in htree.flatten():
            document_id_lst = row.document_id_accumulated
            row['document_id_accumulated'] = document_id_lst
            row['cluster_size'] = len(document_id_lst)
    return htree, n_subclusters


def _print_container(ctr, depth=0):
    """Print summary clustering hierarchy to stdout."""
    message = "[cluster_id={cluster_id}] N_children: {N_children} N_samples: {N_document}"\
              .format(cluster_id=ctr['cluster_id'],
                      N_children=len(ctr.children),
                      N_document=len(ctr['document_id_accumulated']))

    print(''.join(['> ' * depth, message]))

    for child in ctr.children:
        _print_container(child, depth + 1)



# Authors: Roman Yurchak
#
# License: BSD 3 clause

from freediscovery.stop_words import COMMON_FIRST_NAMES, CUSTOM_STOP_WORDS


# Clustering methods for FreeDiscovery
# This is highly inspired from the scikit-learn text clustering example

MAX_N_TOP_WORDS = 1000


def select_top_words(word_list, n=10):
    """ Filter out cluster term names"""
    import re
    from nltk.stem.porter import PorterStemmer
    from sklearn.feature_extraction.stop_words import ENGLISH_STOP_WORDS
    st = PorterStemmer()
    out_st = []
    out = []
    for word in word_list:
        word_st = st.stem(word)
        if len(word_st) <= 2 or\
                re.match('\d+', word_st) or \
                re.match('[^a-zA-Z0-9]', word_st) or\
                word in COMMON_FIRST_NAMES or \
                word in CUSTOM_STOP_WORDS or\
                word in ENGLISH_STOP_WORDS or \
                word_st in out_st:  # ignore stemming duplicate
            continue
        out_st.append(word_st)
        out.append(word)
        if len(out) >= n:
            break
    return out


class _BirchDummy(object):
    """ A dummy class for Birch """
    pass


class ClusterLabels(object):
    """Calculate the cluster labels.


    Parameters
    ----------
    vect : VectorizerMixin object
       a scikit-learn's text vectorizer
    model : ClusterMixin object
       the cluster object
    lsi_components: TruncatedSVD object or None
       LSA object if it was used for clustering
    method: str, optional, default='centroid-frequency'
       the method used to compute the centroid labels
       Only 'centroid-frequency' is supported at the moment.
    n_top_words: int, default=10
       keep only most relevant n_top_words words
    """
    def __init__(self, vect, model, lsi=None,
                 method='centroid-frequency', n_top_words=6):
        self.model = model
        self.vect = vect
        self.lsi = lsi
        self.method = method
        self.n_top_words = n_top_words

    def _to_original_space(self, centroids):
        """If LSI is used recover the data points positions in the original space
        """
        if self.lsi is not None:
            return self.lsi.inverse_transform(centroids)
        else:
            return centroids

    def _get_model_centroids(self):
        method_name = type(self.model).__name__
        if method_name not in ['MiniBatchKMeans', 'AgglomerativeClustering',
                               'Birch', '_BirchDummy', 'DBSCAN']:
            raise NotImplementedError('Method name: '
                                      '{} not implented!'.format(method_name))
        # centroids were previously computed
        return self.model.cluster_centers_

    def predict(self, centroids=None):
        """ Compute the cluster labels

        Parameters
        ----------
        centroids : list, default=None
           if not None, ignore clustering given by the clustering model and
           compute labels for the given cluster centroids

        Returns
        -------
        cluster_labels: array [n_samples]
        """
        if centroids is None:
            centroids = self._get_model_centroids()

        if self.n_top_words > MAX_N_TOP_WORDS:
            raise ValueError
        if self.method == 'centroid-frequency':
            return self._predict_centroid_freq(centroids)
        else:
            raise ValueError

    def _predict_centroid_freq(self, centroids):
        """ Return cluster labels based on
        the most frequent words (tfidf) at cluster centroids """

        centroids = self._to_original_space(centroids)

        n_clusters = centroids.shape[0]

        centroids_ordered = centroids.argsort()[:, ::-1]

        terms = self.vect.get_feature_names()
        cluster_terms = []
        for i in range(n_clusters):
            terms_i = [terms[ind]
                       for ind in centroids_ordered[i, :MAX_N_TOP_WORDS]]
            terms_i = select_top_words(terms_i, self.n_top_words)
            cluster_terms.append(terms_i)
        return cluster_terms



# Authors: Roman Yurchak
#
# License: BSD 3 clause

from __future__ import print_function

import numpy as np


class _DendrogramChildren(object):
    """Compute childen for a given dendogram node

    Parameters
    ----------
    ddata : dict
       data returned by scipy.cluster.hierarchy.dendrogram function
    """
    def __init__(self, ddata):
        self.icoord = np.array(ddata['icoord'])[:, 1:3]
        self.dcoord = np.array(ddata['dcoord'])[:, 1]
        self.icoord_min = self.icoord.min()
        self.icoord_max = self.icoord.max()
        self.leaves = np.array(ddata['leaves'])

    def query(self, node_id):
        """ Get all children for the node (specified by node_id) """
        mask = self.dcoord[node_id] >= self.dcoord

        def _interval_intersect(a0, a1, b0, b1):
            return a0 <= b1 and b0 <= a1

        # essentially intersection of lines from all children nodes
        sort_idx = np.argsort(self.dcoord[mask])[::-1]
        left, right = list(self.icoord[node_id])
        for ileft, iright in self.icoord[mask, :][sort_idx]:
            if _interval_intersect(ileft, iright, left, right):
                left = min(left, ileft)
                right = max(right, iright)

        extent = np.array([left, right])
        extent = (extent - self.icoord_min)*(len(self.leaves) - 1)
        extent /= (self.icoord_max - self.icoord_min)
        extent = extent.astype(int)
        extent = slice(extent[0], extent[1]+1)
        return self.leaves[extent]



# -*- coding: utf-8 -*-

import os.path
from pathlib import Path

import pytest
import numpy as np
import pandas as pd
from numpy.testing import assert_equal, assert_array_equal

from freediscovery.utils import dict2type, sdict_keys
from .base import (parse_res, V01, app, get_features, data_dir,
                   get_features_cached, CACHE_DIR)
from freediscovery.exceptions import (NotFound, WrongParameter)
from sklearn.externals import joblib


# ============================================================================#
#
#                     Feature extraction
#
# ============================================================================#

def test_get_features(app):
    dsid, pars, _ = get_features_cached(app)

    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)
    for key, val in pars.items():
        if key in ['data_dir', 'dataset_definition']:
            continue
        assert val == data[key]

def test_delete_feature_extraction(app):
    dsid, pars, _ = get_features(app)

    method = V01 + "/feature-extraction/{}".format(dsid)
    app.delete_check(method)


def test_get_feature_extraction_all(app):
    method = V01 + "/feature-extraction/"
    data = app.get_check(method)
    for row in data:
        assert sdict_keys(row) == sdict_keys({'analyzer': 'str',
                     'ngram_range': ['int'], 'stop_words': 'NoneType',
                     'n_jobs': 'int', 'chunk_size': 'int',
                     'data_dir': 'str', 'id': 'str', 'n_samples': 'int',
                     'n_features': 'int', 'weighting': 'str', 'norm_alpha': 'float',
                     'use_hashing': 'bool'})


@pytest.mark.parametrize('hashed, weighting', [(True, 'nnc'),
                                             (False, 'nsc'),
                                             (False, 'nnc'),
                                             (True, 'nsc'),
                                             (False, 'lncp'),
                                             (False, 'lnup')])
def test_get_feature_extraction(app, hashed, weighting):
    norm_alpha = 0.5
    dsid, _, _ = get_features_cached(app, hashed=hashed, weighting=weighting,
                                     norm_alpha=norm_alpha)
    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)
    assert dict2type(data, collapse_lists=True) == {'analyzer': 'str',
                     'ngram_range': ['int'], 'stop_words': 'str',
                     'n_jobs': 'int', 'chunk_size': 'int',
                     'data_dir': 'str', 'n_samples': 'int',
                     'n_features': 'int', 'weighting': 'str',
                     'norm_alpha': 'float', 'use_hashing': 'bool',
                     'filenames': ['str'], 'max_df': 'float', 'min_df': 'float',
                     'parse_email_headers': 'bool', 'n_samples_processed': 'int',
                     'preprocess': []}

    assert data['use_hashing'] == hashed
    assert data['weighting'] == weighting
    assert data['norm_alpha'] == norm_alpha

    vect = joblib.load(os.path.join(CACHE_DIR, 'ediscovery_cache', dsid, 'vectorizer'))
    assert (data['use_hashing'] is True) == ('hashing' in type(vect).__name__.lower())


@pytest.mark.parametrize('hashed', [True])
def test_stop_words_integration(app, hashed):
    url = V01 + '/stop-words/'

    sw_name = 'test1w'
    pars = {'name': sw_name,
            'stop_words': ['and', 'or', 'in']}

    res = app.post_check(url, json=pars)
    assert dict2type(res, collapse_lists=True) == {'name': 'str'}
    assert res['name'] == sw_name

    res = app.get_check(url + sw_name)
    assert dict2type(res, collapse_lists=True) == {'name': 'str',
                                                   'stop_words': ['str']}
    assert res['name'] == sw_name
    assert res['stop_words'] == pars['stop_words']

    dsid, pars, _ = get_features(app, hashed=hashed, stop_words=sw_name)



def test_get_search_filenames(app):

    dsid, _, _ = get_features_cached(app)

    method = V01 + "/feature-extraction/{}/id-mapping".format(dsid)

    def _filter_dict(x, filter_field):
        return {key: val for key, val in x.items() if key == filter_field}

    response_ref = {'internal_id': 'int',
                    'file_path': 'str',
                    'document_id': 'int'}

    # Query 1
    file_path_obj = [{'file_path': val} for val in ['00401.txt', '00506.txt']]
    data = app.post_check(method, json={'data': file_path_obj})
    data = data['data']

    for idx in range(len(data)):
        assert dict2type(data[idx]) == response_ref
    assert [_filter_dict(row, 'file_path') for row in data] == file_path_obj
    assert_equal(np.asarray([row['internal_id'] for row in data])**2,
                 [row['document_id'] for row in data])

    with pytest.raises(NotFound):
        res = app.post(method, json={'data': [{'file_path': '00400.txt'}]})

    # Query 2
    file_path_obj = [{'document_id': 4}, {'document_id': 9}]
    data = app.post_check(method, json={'data': file_path_obj})
    data = data['data']

    for idx in range(len(data)):
        assert dict2type(data[idx]) == response_ref
    assert [_filter_dict(row, 'document_id') for row in data] == file_path_obj
    assert_equal(np.asarray([row['internal_id'] for row in data])**2,
                 [row['document_id'] for row in data])


@pytest.mark.parametrize('ingestion_method', ['file_path', 'content'])
def test_append_documents(app, ingestion_method):

    method = V01 + "/feature-extraction/"
    data = app.post_check(method)
    dsid = data['id']
    method += dsid
    app.post_check(method, json={'data_dir': data_dir})

    data = app.get_check(method)

    # check that the file_path is correctly returned by the id-mapping
    data = app.post_check(method + '/id-mapping',
                          json={'return_file_path': False})
    assert dict2type(data['data'][0]) == \
        {'document_id': 'int', 'internal_id': 'int'}

    data = app.post_check(method + '/id-mapping',
                          json={'return_file_path': True})
    assert dict2type(data['data'][0]) == \
        {'document_id': 'int', 'file_path': 'str', 'internal_id': 'int'}
    db_old = data['data']

    dataset_definition = []
    for idx, row in enumerate(db_old):
        row_out = {'document_id': idx + 10}
        if ingestion_method == 'file_path':
            row_out['file_path'] = os.path.join(data_dir, row['file_path'])
        elif ingestion_method == 'content':
            with Path(data_dir, row['file_path']).open('rt', encoding='utf-8') as fh:
                row_out['content'] = fh.read()
        dataset_definition.append(row_out)

    app.post_check(method + '/append',
                   json={'dataset_definition': dataset_definition})
    data = app.post_check(method + '/id-mapping',
                          json={'return_file_path': True})

    db_old = pd.DataFrame(db_old)
    db_new = pd.DataFrame(data['data'])
    assert db_old.shape[0]*2 == db_new.shape[0]


def test_remove_documents(app):
    method = V01 + "/feature-extraction/"
    data = app.post_check(method)
    dsid = data['id']
    method += dsid
    app.post_check(method, json={'data_dir': data_dir})

    data = app.get_check(method)

    data = app.post_check(method + '/id-mapping')
    db_old = pd.DataFrame(data['data'])

    dataset_definition = db_old.iloc[[2, 4]][['document_id']].to_dict(orient='records')
    for row in dataset_definition:
        # convert int64 to int
        row['document_id'] = int(row['document_id'])

    app.post_check(method + '/delete', json={'dataset_definition': dataset_definition})

    data = app.post_check(method + '/id-mapping')

    db_new = pd.DataFrame(data['data'])
    assert db_old.shape[0] - 2 == db_new.shape[0]


@pytest.mark.parametrize('ingestion_method', ['content', 'file_path'])
def test_batch_ingest(app, ingestion_method):
    data_dir1 = Path(data_dir)
    data_dir2 = Path(data_dir).parent.parent / 'ds_002' / 'raw'
    method = V01 + "/feature-extraction/"
    data = app.post_check(method)
    dsid = data['id']
    method += dsid
    for data_dir_k in [data_dir1, data_dir2]:
        # post content from different data dirs
        dd = []
        for fname in data_dir_k.glob('*txt'):
            if ingestion_method == 'content':
                with fname.open('rt', encoding='utf-8') as fh:
                    dd.append({'content': fh.read()})
            elif ingestion_method == 'file_path':
                dd.append({'file_path': str(fname)})
            else:
                raise ValueError('method={} not supported'
                                 .format(ingestion_method))

        app.post_check(method, json={'dataset_definition': dd,
                                     'vectorize': False})
    app.post_check(method, json={'data_dir': str(data_dir1),
                                 'vectorize': False})

    app.post_check(method, json={'vectorize': True})
    data = app.get_check(method)
    assert len(data['filenames']) == 6*3
    assert data['n_samples'] == 6*3


@pytest.mark.parametrize('document_id_generator',
                         [None, 'indexed_file_path', 'infer_file_path'])
def test_document_id_generation(app, document_id_generator):
    method = V01 + "/feature-extraction/"
    data = app.post_check(method)
    dsid = data['id']
    method += dsid
    pars = {'data_dir': data_dir}
    if document_id_generator:
        pars['document_id_generator'] = document_id_generator
    app.post_check(method, json=pars)

    data = app.get_check(method)

    # check that the file_path is correctly returned by the id-mapping
    data = app.post_check(method + '/id-mapping')

    df = pd.DataFrame(data['data'])

    if document_id_generator in [None, 'indexed_file_path']:
        assert_array_equal(df.document_id.values, df.internal_id.values)
    else:
        assert_array_equal(df.document_id.values,
                           [747101442, 747117435, 7628635,
                            7628636, 7628637, 7628638])

def test_document_non_random_id(app):
    method = V01 + "/feature-extraction/"

    dsid_orig = 'test-dir'

    data = app.post_check(method, json={'id': dsid_orig})
    assert data['id'] == dsid_orig

    with pytest.raises(WrongParameter):
        data = app.post_check(method, json={'id': dsid_orig})

    with pytest.raises(WrongParameter):
        data = app.post_check(method, json={'id': 'dsjkdlsy8^$$$'})



# -*- coding: utf-8 -*-

import sys

import pytest
import pandas as pd

from .run_suite import check_cache
from ..utils import dict2type
from unittest import SkipTest

from freediscovery.datasets import load_dataset, IR_DATASETS

cache_dir = check_cache()


@pytest.mark.parametrize('name', ['20_newsgroups_micro'])
def test_load_20newsgoups_dataset(name):
    if (sys.version_info < (3, 0)):
        raise SkipTest
    md, training_set, dataset = load_dataset(name,
                                             cache_dir=cache_dir)

    response_ref = {"document_id": 'int',
                    "file_path": "str",
                    "internal_id": "int"}
    if '20_newsgroups_' in name or 'treclegal09' in name:
        response_ref["category"] = "str"

    assert dict2type(md) == {'data_dir': 'str', 'name': 'str'}

    assert dict2type(dataset[0]) == response_ref
    assert dict2type(training_set[1]) == response_ref

    categories = sorted(list(set([row['category'] for row in dataset])))
    for categories_sel in \
            [[categories[0]],
             [categories[1]],
             [categories[0], categories[1]],
             [categories[1]],
             categories]:

        md, training_set, dataset = load_dataset(name, cache_dir=cache_dir,
                                                 categories=categories_sel)

        for resp in [training_set, dataset]:

            assert dict2type(resp[0]) == response_ref
            result_fields = list(set([el['category'] for el in resp]))

            # the opposite if not always true (e.g. for small training sets)
            for key in result_fields:
                assert key in categories_sel

        training_set = pd.DataFrame(training_set)
        dataset = pd.DataFrame(dataset)
        if name == 'treclegal09_2k_subset':
            if categories_sel == ['positive']:
                assert dataset.shape[0] == 12
            elif categories_sel == categories:
                assert dataset.shape[0] == 2465
                assert (training_set.category == 'positive').sum() == 5
        elif name == '20_newsgroups_micro':
            if categories_sel == ['comp.graphics']:
                assert dataset.shape[0] == 3
            elif categories_sel == categories:
                assert dataset.shape[0] == 7
                assert training_set.shape[0] == 4

# # Uncomment the following section to test the dataset downloading
# #
#
# test_datasets = list(IR_DATASETS.keys())
# test_datasets.remove('legal09int')
#
# @pytest.mark.parametrize('name', test_datasets)
# def test_dataset_downloads(name):
#     load_dataset(name, cache_dir=cache_dir)



"""
.. _exploring_birch_example:

Exploring BIRCH cluster hierarchy
=================================

An example illustrating how to explore the cluster hierarchy
computed by the BIRCH algorithm.

In this example, we use a
:ref:`patched verson <birch_section>` of :class:`sklearn.cluster.Birch`
that allows to store the id of samples belonging to each subcluster.
This modified version is available from :class:`freediscovery.cluster.Birch`.

Building the cluster hierarchy
------------------------------

We start by computing BIRCH clustering on some random structured data,
"""

import numpy as np
from sklearn.datasets import make_blobs
from freediscovery.cluster import Birch, birch_hierarchy_wrapper

rng = np.random.RandomState(42)

X, y = make_blobs(n_samples=1000, n_features=10, random_state=rng)


cluster_model = Birch(threshold=0.9, branching_factor=20,
                      compute_sample_indices=True)
cluster_model.fit(X)

###############################################################################
#
# Next we wrap each subcluster in the cluster hierarchy
# (``cluster_model.root_``) with the
# :class:`~freediscovery.cluster.BirchSubcluster` class
# that allows easier manipulation of the hierarchical tree.

htree, _ = birch_hierarchy_wrapper(cluster_model)
print('Total number of subclusters:', htree.tree_size)


###############################################################################
#
# Visualizing the hierarchy
# -------------------------
# We can now visualize the cluster hierarchy,

htree.display_tree()

###############################################################################
#
# We have a hierarchy 2 levels deep, with 78 sub-clusters and a total
# of 1000 samples.
#
# For instance, let's consider the subcluster with  ``cluster_id=34``.
# We can access it by id with the flattened representation of the hierarchy,

sc = htree.flatten()[34]
print(sc)

###############################################################################
#
# Each subcluster is a dictionary linked inside the hierarchy via the
# `parent` / `children` attributes
# (cf documentation of :class:`~freediscovery.cluster.BirchSubcluster`).
# The ids of the samples contained in a subcluster are stored under the
# ``document_id_accumulated`` key. We can perform any calculations with the
# samples in a given cluster by indexing the original dataset `X`,

print('cluster_id', sc['cluster_id'])
print('document_id_accumulated', sc['document_id_accumulated'])
sc_centroid = X[sc['document_id_accumulated'], :].mean(axis=0)
print(sc_centroid)

###############################################################################
#
# For instance, we can select only subclusters that are one level deep
# (this includes ``cluster_id=34``) and compute their centroids,

htree_depth_1 = [sc for sc in htree.flatten() if sc.current_depth == 1]

for sc in htree_depth_1:
    sc['centroid'] = X[sc['document_id_accumulated'], :].mean(axis=0)

print('Centroid for cluster_id=34:\n', htree.flatten()[34]['centroid'])


###############################################################################
#
# Custom calculations in the hierarchy
# ------------------------------------
#
# While for a number of computations, it is sufficient to iterate through
# a flattened tree, sometimes the output of the calculation need to account
# for data from any number of other subclusters in the tree (e.g. all the
# children). In this case we can subclass
# :class:`~freediscovery.cluster.BirchSubcluster` to add out custom recursive
# function. Here we will add a function that for any subcluster computes the
# the maximum depth of the tree spanned by its children subclusters,

from freediscovery.cluster import BirchSubcluster


class NewBirchSubcluster(BirchSubcluster):

    @property
    def max_tree_depth(self):
        if self.children:
            return 1 + max(sc.max_tree_depth for sc in self.children)
        else:
            return 0


###############################################################################
#
# by re-wrapping the cluster hierarchy with this container, we get,

htree_new, _ = birch_hierarchy_wrapper(cluster_model,
                                       container=NewBirchSubcluster)

print('Tree depth from the root node:', htree_new.max_tree_depth)

print('Tree depth from cluster_id=34:', htree_new.flatten()[34].max_tree_depth)



# -*- coding: utf-8 -*-
#
# Authors: Roman Yurchak
#
# License: BSD 3 clause

import warnings

import scipy.sparse as sp
import numpy as np

from sklearn.utils.validation import check_array
from sklearn.preprocessing import normalize
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.sparsefuncs_fast import csr_row_norms
from sklearn.exceptions import DataConversionWarning

from .utils import _mean_csr_nonzero_axis1


def _document_frequency(X):
    """Count the number of non-zero values for each feature in sparse X.
    (copied from scikit-learn)
    """
    if sp.isspmatrix_csr(X):
        return np.bincount(X.indices, minlength=X.shape[1])
    else:
        return np.diff(sp.csc_matrix(X, copy=False).indptr)


def _document_length(X):
    return X.sum(axis=1)


def _validate_smart_notation(scheme):

    if not isinstance(scheme, str) or len(scheme) not in [3, 4]:
        raise ValueError('Expected a 3 or 4 character long string for scheme, '
                         'got {}'.format(scheme))

    if len(scheme) == 3:
        scheme_t, scheme_d, scheme_n = scheme
    elif len(scheme) == 4:
        scheme_t, scheme_d, scheme_n1, scheme_n2 = scheme
        scheme_n = ''.join([scheme_n1, scheme_n2])

    if scheme_t not in 'nlabLd':
        raise ValueError(('Term frequency weighting {} '
                          'not supported, must be one of nlabL')
                         .format(scheme_t))
    if scheme_d not in 'ntspd':
        raise ValueError(('Document frequency weighting {} '
                          'not supported, must be one of ntp')
                         .format(scheme_d))
    if scheme_n not in ['n', 'c', 'l', 'u', 'cp', 'lp', 'up', 'b']:
        raise ValueError(('Document normalization {} '
                          'not supported, must be of the form [nclub][p]?')
                         .format(scheme_n))
    if scheme_n not in ['n', 'c', 'l', 'u', 'cp', 'lp', 'up', ]:
        raise NotImplementedError(
                   ('Document normalization {}'
                    'is not yet implemented, must be of the form [nclu][p]?')
                   .format(scheme_n))
    return scheme_t, scheme_d, scheme_n


class SmartTfidfTransformer(BaseEstimator, TransformerMixin):
    """TF-IDF weighting and normalization with the SMART IR notation

    This class is similar to
    :class:`sklearn.feature_extraction.text.TfidfTransformer` but supports
    a larger number of TF-IDF weighting and normalization schemes.
    It should be fitted on the document-term matrix computed by
    :class:`sklearn.feature_extraction.text.CountVectorizer`.

    The TF-IDF transform consists of three subsequent operations, determined
    by the ``weighting`` parameter,

    1. Term frequency weighing:

       natural (``n``), log (``l``), augmented
       (``a``),  boolean (``b``), log average (``L``)

    2. Document frequency weighting:

       none (``n``), idf (``t``), smoothed
       idf (``s``),  probabilistic (``p``), smoothed probabilistic (``d``)

    3. Document normalization:

       none (``n``), cosine (``c``), length (``l``),
       unique (``u``).

    Following the SMART IR notation, the ``weighting`` parameter is written
    as the concatenation of thee characters describing each processing step.
    In addition the pivoted normalization can be enabled with a fourth
    character ``p``.


    See the :ref:`tfidf_section` documentation section for more details.


    Parameters
    ----------
    weighting : str, default='nsc'
      the SMART notation for document, term weighting and normalization.
      In the form ``[nlabL][ntspd][ncb][p]``.
    norm_alpha : float, default=0.75
      the α parameter in the pivoted normalization. This parameter is only
      used when ``weighting='???p'``.
    norm_pivot : float, default=None
      the pivot value used for the normalization. If not provided
      it is computed as the mean of the ``norm(tf*idf)``. This parameter is
      only used when ``weighting='???p'``.
    compute_df : bool, default=False
      compute the document frequency (``df_`` attribute) even when it's not
      explicitly required by the weighting scheme.
    copy : boolean, default=True
      Whether to copy the input array and operate on the copy or perform
      in-place operations in fit and transform.


    See also
    --------
    :class:`sklearn.feature_extraction.text.TfidfTransformer`


    References
    ----------
    .. [Manning2008] C.D. Manning, P. Raghavan, H. Schütze,
       `"Document and query weighting schemes"
       <https://nlp.stanford.edu/IR-book/html/htmledition/document-and-query-weighting-schemes-1.html>`_ ,
       2008
    .. [Singhal1996] A. Singhal, C. Buckley, and M. Mitra.
       `"Pivoted document length normalization."
       <https://ecommons.cornell.edu/bitstream/handle/1813/7217/95-1560.pdf?sequence=1>`_ , 1996
    """ # noqa
    def __init__(self, weighting='nsc', norm_alpha=0.75, norm_pivot=None,
                 compute_df=False, copy=True):
        _validate_smart_notation(weighting)
        self.weighting = weighting
        self.norm_alpha = norm_alpha
        self.norm_pivot = norm_pivot
        self.compute_df = compute_df
        self.copy = copy

    def fit(self, X, y=None):
        """Learn the document lenght and document frequency vector
        (if necessary).

        Parameters
        ----------
        X : sparse matrix, [n_samples, n_features]
            a matrix of term/token counts
        """
        X = check_array(X, ['csr'], copy=self.copy)
        scheme_t, scheme_d, scheme_n = _validate_smart_notation(self.weighting)
        self.dl_ = _document_length(X)
        if scheme_d in 'stp' or self.compute_df:
            self.df_ = _document_frequency(X)
        else:
            self.df_ = None
        if sp.isspmatrix_csr(X):
            self.du_ = np.diff(X.indptr)
        else:
            self.du_ = X.shape[-1] - (X == 0).sum(axis=1)
        self._n_features = X.shape[1]

        if self.df_ is not None:
            df_n_samples = len(self.dl_)
        else:
            df_n_samples = None

        if scheme_n.endswith('p') and self.norm_pivot is None:
            # Need to compute the pivot if it's not provided
            _, self.norm_pivot = _smart_tfidf(X, self.weighting, self.df_,
                                              df_n_samples,
                                              norm_alpha=self.norm_alpha,
                                              norm_pivot=self.norm_pivot,
                                              return_pivot=True)

        return self

    def fit_transform(self, X, y=None):
        """Apply document term weighting and normalization on text features

        Parameters
        ----------
        X : sparse matrix, [n_samples, n_features]
            a matrix of term/token counts
        """
        X = check_array(X, ['csr'], copy=self.copy)

        scheme_t, scheme_d, scheme_n = _validate_smart_notation(self.weighting)
        self.dl_ = _document_length(X)
        if scheme_d in 'stpd' or self.compute_df:
            self.df_ = _document_frequency(X)
        else:
            self.df_ = None
        if sp.isspmatrix_csr(X):
            self.du_ = np.diff(X.indptr)
        else:
            self.du_ = X.shape[-1] - (X == 0).sum(axis=1)
        self._n_features = X.shape[1]

        if self.df_ is not None:
            df_n_samples = len(self.dl_)
        else:
            df_n_samples = None

        if self.df_ is not None:
            df_n_samples = len(self.dl_)
        else:
            df_n_samples = None

        X, self.norm_pivot = _smart_tfidf(X, self.weighting, self.df_,
                                          df_n_samples,
                                          norm_alpha=self.norm_alpha,
                                          norm_pivot=self.norm_pivot,
                                          return_pivot=True)
        return X

    def transform(self, X, y=None):
        """Apply document term weighting and normalization on text features

        Parameters
        ----------
        X : sparse matrix, [n_samples, n_features]
            a matrix of term/token counts
        copy : boolean, default True
            Whether to copy X and operate on the copy or perform in-place
            operations.
        """
        X = check_array(X, ['csr'], copy=self.copy)
        check_is_fitted(self, 'dl_', 'vector is not fitted')
        if X.shape[1] != self._n_features:
            raise ValueError(('Model fitted with n_features={} '
                              'but X.shape={}')
                             .format(self._n_features, X.shape))

        if self.df_ is not None:
            df_n_samples = len(self.dl_)
        else:
            df_n_samples = None

        return _smart_tfidf(X, self.weighting, self.df_,
                            df_n_samples,
                            norm_alpha=self.norm_alpha,
                            norm_pivot=self.norm_pivot)


def _smart_tfidf(tf, weighting, df=None, df_n_samples=None, norm_alpha=0.75,
                 norm_pivot=None, return_pivot=False):
    """
    Apply TF-IDF feature weighting using the SMART notation.


    Parameters
    ----------
    df : sparse csr array
      the term frequency matrix (n_documents, n_features)

    weighting : str, default='nnc'
      the SMART notation for document term weighting and normalization.
      In the form [nlabL][ntspd][nclu][p] , see
      https://en.wikipedia.org/wiki/SMART_Information_Retrieval_System

    df : array, shape=[n_features], optional
      precomputed inverse document frequency matrix (n_features,).
      If not provided, it will be recomputed if necessary. Both df and
      df_n must be provided at the same time.

    df_n_samples : float, default=None
      when using a inverse document frequency matrix, the number of
      documents that were used to compute the df. Both df and df_n
      must be provided at the same time.

    norm_alpha : float, default=0.75
      the alpha parameter in the pivoted normalization. Only used when
      weighting='???p'.

    norm_pivot : float, default=None
      the pivot value used for the normalization. If not provided, and
      weighting='???p', it is computed as the mean of the norm(tf*idf).

    return_pivot : bool, default=False
      return the computed norm_pivot

    Returns
    -------

    X : sparse csr array
      the weighted term frequency matrix

    norm_pivot : flot
      return the norm pivot (only when return_pivot=True)

    References
    ----------
    .. [Manning2008] C.D. Manning, P. Raghavan, H. Schütze,
       `"Document and query weighting schemes"
       <https://nlp.stanford.edu/IR-book/html/htmledition/document-and-query-weighting-schemes-1.html>`_ ,
       2008
    .. [Singhal1996] A. Singhal, C. Buckley, and M. Mitra.
       `"Pivoted document length normalization."
       <https://ecommons.cornell.edu/bitstream/handle/1813/7217/95-1560.pdf?sequence=1>`_ , 1996
    """  # noqa

    tf = check_array(tf, ['csr'])
    if (df is None) != (df_n_samples is None):
        raise ValueError(('df={} and df_n_samples={}, while both should be '
                          'either provided or not provided')
                         .format(df is None, df_n_samples))
    if df is not None:
        df = check_array(df, ensure_2d=False)
        if df.shape[0] != tf.shape[-1]:
            raise ValueError(('df array provided with n_features={} ,'
                              'while in the tf array n_features={}')
                             .format(df.shape[0], tf.shape[1]))

    if not 0 <= norm_alpha <= 1:
        raise ValueError('norm_alpha={} not in [0, 1]'.format(norm_alpha))

    n_samples, n_features = tf.shape
    if df_n_samples is None:
        df_n_samples = n_samples

    scheme_t, scheme_d, scheme_n = _validate_smart_notation(weighting)

    X = tf

    # term weighting
    if scheme_t == 'n':
        pass
    elif scheme_t == 'l':
        X.data = 1 + np.log(tf.data)
    elif scheme_t == 'd':
        X.data = 1 + np.log(1 + np.log(tf.data))
    elif scheme_t == 'a':
        max_tf = np.squeeze(tf.max(axis=1).A)
        # if max_tf is zero, the tf are going to be all zero anyway
        # so we set it to 1 in order to prevent overflows
        max_tf[max_tf == 0] = 1
        _max_tf_diag = sp.spdiags(1. / max_tf, diags=0, m=n_samples,
                                  n=n_samples, format='csr')
        X = 0.5 * _max_tf_diag.dot(tf)
        X.data += 0.5

    elif scheme_t == 'b':
        X.data = tf.data.astype('bool').astype('int')
    elif scheme_t == 'L':
        mean_tf = _mean_csr_nonzero_axis1(tf)
        # if mean_tf is zero, the tf are going to be all zero anyway
        # so we set it to 1 in order to prevent overflows
        mean_tf[mean_tf == 0] = 1.0
        mean_tf = (1 + np.log(mean_tf))
        _mean_tf_diag = sp.spdiags(1./mean_tf, diags=0, m=n_samples,
                                   n=n_samples, format='csr')

        X.data = (1 + np.log(tf.data))
        X = _mean_tf_diag.dot(X)
    else:
        raise ValueError

    # document weighting
    if scheme_d == 'n':
        pass
    elif scheme_d in 'tpsd':
        if df is None:
            df = _document_frequency(tf)
        if scheme_d == 't':
            idf = np.log(float(df_n_samples) / df) + 1.0
        elif scheme_d == 's':
            idf = np.log(float(df_n_samples + 1) / (df + 1)) + 1.0
        elif scheme_d == 'p':
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore",
                                        message="divide by zero encountered in log",  # noqa
                                        category=RuntimeWarning)
                idf = np.log((float(df_n_samples) - df)/df)
        elif scheme_d == 'd':
            idf = np.log((float(df_n_samples) + 1 - df)/(df + 1))
        _idf_diag = sp.spdiags(idf, diags=0, m=n_features,
                               n=n_features, format='csr')
        X = X.dot(_idf_diag)
    else:
        raise ValueError

    # normalization
    if scheme_n == 'n':
        pass
    elif scheme_n == 'c':
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",
                                    category=DataConversionWarning)
            X = normalize(X, norm="l2", copy=False)
    elif scheme_n == 'l':
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",
                                    category=DataConversionWarning)
            X = normalize(X, norm="l1", copy=False)
    elif scheme_n == 'u':
        X_norm = np.diff(X.indptr)
        X_norm[X_norm == 0] = 1.
        # empty documents (with a zero norm) don't need to be normalized
        _diag_norm = sp.spdiags(1./X_norm, diags=0, m=n_samples,
                                n=n_samples, format='csr')
        X = _diag_norm.dot(X)
    elif scheme_n in ['cp', 'lp', 'up']:
        if scheme_n == 'cp':
            X_norm = np.sqrt(csr_row_norms(X))
        elif scheme_n == 'lp':
            X_data = X.data.copy()
            X.data = np.abs(X.data)
            X_norm = np.squeeze(X.sum(axis=1).A)
            X.data = X_data
        elif scheme_n == 'up':
            X_norm = np.diff(X.indptr)

        if norm_pivot is None:
            norm_pivot = X_norm.mean()

        # empty documents (with a zero norm) don't need to be normalized
        X_norm[X_norm == 0] = 1.

        pivoted_norm = (1 - norm_alpha)*norm_pivot + norm_alpha*X_norm
        _diag_pivoted_norm = sp.spdiags(1./pivoted_norm, diags=0, m=n_samples,
                                        n=n_samples, format='csr')
        X = _diag_pivoted_norm.dot(X)
    else:
        raise ValueError
    if return_pivot:
        return X, norm_pivot
    else:
        return X



# -*- coding: utf-8 -*-

import numpy as np
from sklearn.externals import joblib

from freediscovery.engine.base import _BaseWrapper
from freediscovery.lsi import _compute_lsi_dimensionality, _TruncatedSVD_LSI
from freediscovery.utils import setup_model
from freediscovery.exceptions import WrongParameter


class _LSIWrapper(_BaseWrapper):
    """Document categorization using Latent Semantic Indexing (LSI)

    Parameters
    ----------
    cache_dir : str
       folder where the model will be saved
    parent_id : str
       dataset id
    mid : str
       LSI model id (the dataset id will be inferred)
    verbose : bool, optional
       print progress messages
    """

    _wrapper_type = "lsi"

    def __init__(self, cache_dir='/tmp/', parent_id=None,
                 mid=None, verbose=False, random_state=None, mode='r'):

        if mode not in ['r', 'w', 'fw']:
            raise WrongParameter('mode={} must be one of "r", "w", "fw"'
                                 .format(mode))
        self.mode = mode

        super(_LSIWrapper, self).__init__(cache_dir=cache_dir,
                                          parent_id=parent_id, mid=mid,
                                          mode=mode)
        self.random_state = random_state

    def _load_features(self):
        mid_dir = self.fe.dsid_dir / self._wrapper_type / self.mid
        return joblib.load(str(mid_dir / 'data'))

    def fit_transform(self, n_components=150, n_iter=5, alpha=0.33):
        """
        Perform the SVD decomposition

        Parameters
        ----------
        n_components : int
           number of selected singular values (number of LSI dimensions)
        n_iter : int
           number of iterations for the stochastic SVD algorithm

        Returns
        -------
        mid : str
           model id
        lsi : _BaseWrapper
           the TruncatedSVD object
        exp_var : float
           the explained variance of the SVD decomposition
        """
        parent_id = self.pipeline.mid

        dsid_dir = self.fe.dsid_dir
        if not dsid_dir.exists():
            raise IOError

        pars = {'parent_id': parent_id, 'n_components': n_components}

        mid_dir_base = dsid_dir / self._wrapper_type

        mid, mid_dir = setup_model(mid_dir_base, mid=self.mid, mode=self.mode)

        ds = self.pipeline.data
        n_components_opt = _compute_lsi_dimensionality(n_components, *ds.shape,
                                                       alpha=alpha)
        svd = _TruncatedSVD_LSI(n_components=n_components_opt,
                                n_iter=n_iter, random_state=self.random_state)
        lsi = svd
        lsi.fit(ds)

        ds_p = lsi.transform_lsi_norm(ds)

        joblib.dump(pars, str(mid_dir / 'pars'))
        joblib.dump(lsi, str(mid_dir / 'model'))
        joblib.dump(ds_p, str(mid_dir / 'data'))

        exp_var = lsi.explained_variance_ratio_.sum()
        self.mid = mid

        return lsi, exp_var

    def append(self, X):
        """ Add new documents to an existing LSI model

        Parameters
        ----------
        X : scipy.sparse CSR array
          the additional data to add to the index

        """
        mid_dir = self.model_dir / self.mid
        lsi = joblib.load(str(mid_dir / 'model'))
        Y_old = joblib.load(str(mid_dir / 'data'))
        Y_new = lsi.transform_lsi_norm(X)
        Y = np.vstack((Y_old, Y_new))
        joblib.dump(Y, str(mid_dir / 'data'))



import re
from ..exceptions import WrongParameter


def validate_mid(mid):
    """Validate a user provided dataset id"""

    if not re.match('^[a-zA-Z0-9_\-]+$', mid):
        raise WrongParameter(('id={} is not valid. '
                              'It can only contain letters, numbers '
                              'and "-", "_" characters. ')
                             .format(mid))

    if len(mid) < 2 or len(mid) > 50:
        raise WrongParameter(('id={} is not valid. '
                              'It must be between 2 and 50 characters long.')
                             .format(mid))



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

def _is_inside_docker():
    """ An imperfect way of checking that the server is
    running inside a Docker container"""

    return os.path.exists('/freediscovery_shared/')




# Authors: Roman Yurchak
#
# License: BSD 3 clause

import numpy as np


def _sample_tree(node, min_similarity):
    """Perform the first step of optimal sampling"""
    from itertools import chain

    if node['cluster_similarity'] > min_similarity or not node.children:
        return [node]
    else:
        return list(chain.from_iterable(
                    _sample_tree(child, min_similarity)
                    for child in node.children))


def compute_optimal_sampling(htree, min_similarity, min_coverage):
    """
    Given a Birch hierarchical tree, compute the optmal sampling,
    with the following steps

      1. For each node, walk through subclusters, and append those
         that have a cluster_similarity > min_similarity,
         recursively iterate through children until
         either this condition is reached or maximum depth is reached.
      2. Sort obtained subclusters by size, and select
         the first N veryfing the coverage requirement
      3. For each resulting subcluster, keep only the document closest
         to the centroid remove children, and the document.

    .. warning::
        This function does modify the original htree
    """
    if not (0 <= min_coverage <= 1):
        raise ValueError('min_coverage={} must be in [0, 1]'.format(
                         min_coverage))

    total_documents = htree['cluster_size']  # root node
    limit_size = total_documents*min_coverage
    # Step 1
    stree = _sample_tree(htree, min_similarity)

    # Step 2
    stree2 = []
    cum_size = 0
    for row in sorted(stree, key=lambda x: x['cluster_size'], reverse=True):
        cum_size += row['cluster_size']
        # Step 3:

        sampled_doc_idx = np.argmax(row['document_similarity'])
        row['document_similarity'] = [row['document_similarity'][sampled_doc_idx]]
        row['document_id_accumulated'] = [row['document_id_accumulated'][sampled_doc_idx]]

        stree2.append(row)
        if cum_size > limit_size:
            # required coverage reached
            break

    return stree2
